#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2014-2018 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""GOES HRIT format reader.

References:
      LRIT/HRIT Mission Specific Implementation, February 2012
      GVARRDL98.pdf
      05057_SPE_MSG_LRIT_HRI

"""

import logging
from datetime import datetime, timedelta

import dask.array as da
import numpy as np
import xarray as xr
from pyresample import geometry

from satpy.readers.eum_base import recarray2dict, time_cds_short
from satpy.readers.hrit_base import (
    HRITFileHandler,
    ancillary_text,
    annotation_header,
    base_hdr_map,
    image_data_function,
)


class CalibrationError(Exception):
    """Dummy error-class."""


logger = logging.getLogger('hrit_goes')

# Geometric constants [meters]
EQUATOR_RADIUS = 6378169.00
POLE_RADIUS = 6356583.80
ALTITUDE = 35785831.00

# goes implementation:
key_header = np.dtype([('key_number', 'u1'),
                       ('seed', '>f8')])

segment_identification = np.dtype([('GP_SC_ID', '>i2'),
                                   ('spectral_channel_id', '>i1'),
                                   ('segment_sequence_number', '>u2'),
                                   ('planned_start_segment_number', '>u2'),
                                   ('planned_end_segment_number', '>u2'),
                                   ('data_field_representation', '>i1')])

image_segment_line_quality = np.dtype([('line_number_in_grid', '>i4'),
                                       ('line_mean_acquisition',
                                        [('days', '>u2'),
                                         ('milliseconds', '>u4')]),
                                       ('line_validity', 'u1'),
                                       ('line_radiometric_quality', 'u1'),
                                       ('line_geometric_quality', 'u1')])

goms_variable_length_headers = {
    image_segment_line_quality: 'image_segment_line_quality'}

goms_text_headers = {image_data_function: 'image_data_function',
                     annotation_header: 'annotation_header',
                     ancillary_text: 'ancillary_text'}

goes_hdr_map = base_hdr_map.copy()
goes_hdr_map.update({7: key_header,
                     128: segment_identification,
                     129: image_segment_line_quality
                     })


orbit_coef = np.dtype([('StartTime', time_cds_short),
                       ('EndTime', time_cds_short),
                       ('X', '>f8', (8, )),
                       ('Y', '>f8', (8, )),
                       ('Z', '>f8', (8, )),
                       ('VX', '>f8', (8, )),
                       ('VY', '>f8', (8, )),
                       ('VZ', '>f8', (8, ))])

attitude_coef = np.dtype([('StartTime', time_cds_short),
                          ('EndTime', time_cds_short),
                          ('XofSpinAxis', '>f8', (8, )),
                          ('YofSpinAxis', '>f8', (8, )),
                          ('ZofSpinAxis', '>f8', (8, ))])

cuc_time = np.dtype([('coarse', 'u1', (4, )),
                     ('fine', 'u1', (3, ))])


sgs_time = np.dtype([('century', 'u1'),
                     ('year', 'u1'),
                     ('doy1', 'u1'),
                     ('doy_hours', 'u1'),
                     ('hours_mins', 'u1'),
                     ('mins_secs', 'u1'),
                     ('secs_msecs', 'u1'),
                     ('msecs', 'u1')])


def make_sgs_time(sgs_time_array):
    """Make sgs time."""
    year = ((sgs_time_array['century'] >> 4) * 1000 +
            (sgs_time_array['century'] & 15) * 100 +
            (sgs_time_array['year'] >> 4) * 10 +
            (sgs_time_array['year'] & 15))
    doy = ((sgs_time_array['doy1'] >> 4) * 100 +
           (sgs_time_array['doy1'] & 15) * 10 +
           (sgs_time_array['doy_hours'] >> 4))
    hours = ((sgs_time_array['doy_hours'] & 15) * 10 +
             (sgs_time_array['hours_mins'] >> 4))
    mins = ((sgs_time_array['hours_mins'] & 15) * 10 +
            (sgs_time_array['mins_secs'] >> 4))
    secs = ((sgs_time_array['mins_secs'] & 15) * 10 +
            (sgs_time_array['secs_msecs'] >> 4))
    msecs = ((sgs_time_array['secs_msecs'] & 15) * 100 +
             (sgs_time_array['msecs'] >> 4) * 10 +
             (sgs_time_array['msecs'] & 15))
    return (datetime(int(year), 1, 1) +
            timedelta(days=int(doy - 1),
                      hours=int(hours),
                      minutes=int(mins),
                      seconds=int(secs),
                      milliseconds=int(msecs)))


satellite_status = np.dtype([("TagType", "<u4"),
                             ("TagLength", "<u4"),
                             ("SatelliteID", "<u8"),
                             ("SatelliteName", "S256"),
                             ("NominalLongitude", "<f8"),
                             ("SatelliteCondition", "<u4"),
                             ("TimeOffset", "<f8")])

image_acquisition = np.dtype([("TagType", "<u4"),
                              ("TagLength", "<u4"),
                              ("Status", "<u4"),
                              ("StartDelay", "<i4"),
                              ("Cel", "<f8")])


gvar_float = '>i4'


def make_gvar_float(float_val):
    """Make gvar float."""
    sign = 1
    if float_val < 0:
        float_val = -float_val
        sign = -1

    exp = (float_val >> 24) - 64
    mant = float_val & ((1 << 24) - 1)
    if mant == 0:
        return 0.
    res = sign * mant * 2.0**(-24 + exp * 4)
    return res


prologue = np.dtype([
    # common generic header
    ("CommonHeaderVersion", "u1"),
    ("Junk1", "u1", 3),
    ("NominalSGSProductTime", time_cds_short),
    ("SGSProductQuality", "u1"),
    ("SGSProductCompleteness", "u1"),
    ("SGSProductTimeliness", "u1"),
    ("SGSProcessingInstanceId", "u1"),
    ("BaseAlgorithmVersion", "S1", 16),
    ("ProductAlgorithmVersion", "S1", 16),
    # product header
    ("ImageProductHeaderVersion", "u1"),
    ("Junk2", "u1", 3),
    ("ImageProductHeaderLength", ">u4"),
    ("ImageProductVersion", "u1"),
    # first block-0
    ("SatelliteID", "u1"),
    ("SPSID", "u1"),
    ("IScan", "u1", 4),
    ("IDSub", "u1", 16),
    ("TCurr", sgs_time),
    ("TCHED", sgs_time),
    ("TCTRL", sgs_time),
    ("TLHED", sgs_time),
    ("TLTRL", sgs_time),
    ("TIPFS", sgs_time),
    ("TINFS", sgs_time),
    ("TISPC", sgs_time),
    ("TIECL", sgs_time),
    ("TIBBC", sgs_time),
    ("TISTR", sgs_time),
    ("TLRAN", sgs_time),
    ("TIIRT", sgs_time),
    ("TIVIT", sgs_time),
    ("TCLMT", sgs_time),
    ("TIONA", sgs_time),
    ("RelativeScanCount", '>u2'),
    ("AbsoluteScanCount", '>u2'),
    ("NorthernmostScanLine", '>u2'),
    ("WesternmostPixel", '>u2'),
    ("EasternmostPixel", '>u2'),
    ("NorthernmostFrameLine", '>u2'),
    ("SouthernmostFrameLine", '>u2'),
    ("0Pixel", '>u2'),
    ("0ScanLine", '>u2'),
    ("0Scan", '>u2'),
    ("SubSatScan", '>u2'),
    ("SubSatPixel", '>u2'),
    ("SubSatLatitude", gvar_float),
    ("SubSatLongitude", gvar_float),
    ("Junk4", "u1", 96),  # move to "word" 295
    ("IMCIdentifier", "S4"),
    ("Zeros", "u1", 12),
    ("ReferenceLongitude", gvar_float),
    ("ReferenceDistance",  gvar_float),
    ("ReferenceLatitude",  gvar_float)
])


class HRITGOESPrologueFileHandler(HRITFileHandler):
    """GOES HRIT format reader."""

    def __init__(self, filename, filename_info, filetype_info):
        """Initialize the reader."""
        super(HRITGOESPrologueFileHandler, self).__init__(filename, filename_info,
                                                          filetype_info,
                                                          (goes_hdr_map,
                                                           goms_variable_length_headers,
                                                           goms_text_headers))
        self.prologue = {}
        self.read_prologue()

    def read_prologue(self):
        """Read the prologue metadata."""
        with open(self.filename, "rb") as fp_:
            fp_.seek(self.mda['total_header_length'])
            data = np.fromfile(fp_, dtype=prologue, count=1)
            self.prologue.update(recarray2dict(data))

        self.process_prologue()

    def process_prologue(self):
        """Reprocess prologue to correct types."""
        for key in ['TCurr', 'TCHED', 'TCTRL', 'TLHED', 'TLTRL', 'TIPFS',
                    'TINFS', 'TISPC', 'TIECL', 'TIBBC', 'TISTR', 'TLRAN',
                    'TIIRT', 'TIVIT', 'TCLMT', 'TIONA']:
            try:
                self.prologue[key] = make_sgs_time(self.prologue[key])
            except ValueError:
                self.prologue.pop(key, None)
                logger.debug("Invalid data for %s", key)

        for key in ['SubSatLatitude', "SubSatLongitude", "ReferenceLongitude",
                    "ReferenceDistance", "ReferenceLatitude"]:
            self.prologue[key] = make_gvar_float(self.prologue[key])


radiometric_processing = np.dtype([("TagType", "<u4"),
                                   ("TagLength", "<u4"),
                                   ("RPSummary",
                                    [("Impulse", "<u4"),
                                     ("IsStrNoiseCorrection", "<u4"),
                                        ("IsOptic", "<u4"),
                                        ("IsBrightnessAligment", "<u4")]),
                                   ("OpticCorrection",
                                    [("Degree", "<i4"),
                                     ("A", "<f8", (16, ))]),
                                   ("RPQuality",
                                    [("EffDinRange", "<f8"),
                                     ("EathDarkening", "<f8"),
                                        ("Zone", "<f8"),
                                        ("Impulse", "<f8"),
                                        ("Group", "<f8"),
                                        ("DefectCount", "<u4"),
                                        ("DefectProcent", "<f8"),
                                        ("S_Noise_DT_Preflight", "<f8"),
                                        ("S_Noise_DT_Bort", "<f8"),
                                        ("S_Noise_DT_Video", "<f8"),
                                        ("S_Noise_DT_1_5", "<f8"),
                                        ("CalibrStability", "<f8"),
                                        ("TemnSKO", "<f8", (2, )),
                                        ("StructSKO", "<f8", (2, )),
                                        ("Struct_1_5", "<f8"),
                                        ("Zone_1_ 5", "<f8"),
                                        ("RadDif", "<f8")])])

geometric_processing = np.dtype([("TagType", "<u4"),
                                 ("TagLength", "<u4"),
                                 ("TGeomNormInfo",
                                  [("IsExist", "<u4"),
                                   ("IsNorm", "<u4"),
                                      ("SubLon", "<f8"),
                                      ("TypeProjection", "<u4"),
                                      ("PixInfo", "<f8", (4, ))]),
                                 ("SatInfo",
                                  [("TISO",
                                    [("T0", "<f8"),
                                     ("dT", "<f8"),
                                        ("ASb", "<f8"),
                                        ("Evsk", "<f8", (3, 3, 4)),
                                        ("ARx", "<f8", (4, )),
                                        ("ARy", "<f8", (4, )),
                                        ("ARz", "<f8", (4, )),
                                        ("AVx", "<f8", (4, )),
                                        ("AVy", "<f8", (4, )),
                                        ("AVz", "<f8", (4, ))]),
                                      ("Type", "<i4")]),
                                 ("TimeProcessing", "<f8"),
                                 ("ApriorAccuracy", "<f8"),
                                 ("RelativeAccuracy", "<f8", (2, ))])


C1 = 1.19104273e-5
C2 = 1.43877523

SPACECRAFTS = {
    # these are GP_SC_ID
    18007: "GOES-7",
    18008: "GOES-8",
    18009: "GOES-9",
    18010: "GOES-10",
    18011: "GOES-11",
    18012: "GOES-12",
    18013: "GOES-13",
    18014: "GOES-14",
    18015: "GOES-15",
    # these are in block-0
    7: "GOES-7",
    8: "GOES-8",
    9: "GOES-9",
    10: "GOES-10",
    11: "GOES-11",
    12: "GOES-12",
    13: "GOES-13",
    14: "GOES-14",
    15: "GOES-15"}


class HRITGOESFileHandler(HRITFileHandler):
    """GOES HRIT format reader."""

    def __init__(self, filename, filename_info, filetype_info,
                 prologue):
        """Initialize the reader."""
        super(HRITGOESFileHandler, self).__init__(filename, filename_info,
                                                  filetype_info,
                                                  (goes_hdr_map,
                                                   goms_variable_length_headers,
                                                   goms_text_headers))
        self.prologue = prologue.prologue
        self.chid = self.mda['spectral_channel_id']

        sublon = self.prologue['SubSatLongitude']
        self.mda['projection_parameters']['SSP_longitude'] = sublon

        satellite_id = self.prologue['SatelliteID']
        self.platform_name = SPACECRAFTS[satellite_id]

    def get_dataset(self, key, info):
        """Get the data  from the files."""
        logger.debug("Getting raw data")
        res = super(HRITGOESFileHandler, self).get_dataset(key, info)

        self.mda['calibration_parameters'] = self._get_calibration_params()

        res = self.calibrate(res, key['calibration'])
        new_attrs = info.copy()
        new_attrs.update(res.attrs)
        res.attrs = new_attrs
        res.attrs['platform_name'] = self.platform_name
        res.attrs['sensor'] = 'goes_imager'
        res.attrs['orbital_parameters'] = {'projection_longitude': self.mda['projection_parameters']['SSP_longitude'],
                                           'projection_latitude': 0.0,
                                           'projection_altitude': ALTITUDE}
        return res

    def _get_calibration_params(self):
        """Get the calibration parameters from the metadata."""
        params = {}
        idx_table = []
        val_table = []
        for elt in self.mda['image_data_function'].split(b'\r\n'):
            try:
                key, val = elt.split(b':=')
                try:
                    idx_table.append(int(key))
                    val_table.append(float(val))
                except ValueError:
                    params[key] = val
            except ValueError:
                pass
        params['indices'] = np.array(idx_table)
        params['values'] = np.array(val_table, dtype=np.float32)
        return params

    def calibrate(self, data, calibration):
        """Calibrate the data."""
        logger.debug("Calibration")
        tic = datetime.now()
        if calibration == 'counts':
            return data
        if calibration == 'reflectance':
            res = self._calibrate(data)
        elif calibration == 'brightness_temperature':
            res = self._calibrate(data)
        else:
            raise NotImplementedError("Don't know how to calibrate to " +
                                      str(calibration))

        logger.debug("Calibration time " + str(datetime.now() - tic))
        return res

    def _calibrate(self, data):
        """Calibrate *data*."""
        idx = self.mda['calibration_parameters']['indices']
        val = self.mda['calibration_parameters']['values']
        data.data = da.where(data.data == 0, np.nan, data.data)
        ddata = data.data.map_blocks(np.interp, idx, val, dtype=val.dtype)
        res = xr.DataArray(ddata,
                           dims=data.dims, attrs=data.attrs,
                           coords=data.coords)
        res = res.clip(min=0)
        units = {b'percent': '%', b'degree Kelvin': 'K'}
        unit = self.mda['calibration_parameters'][b'_UNIT']
        res.attrs['units'] = units.get(unit, unit)
        return res

    def get_area_def(self, dsid):
        """Get the area definition of the band."""
        cfac = np.int32(self.mda['cfac'])
        lfac = np.int32(self.mda['lfac'])
        coff = np.float32(self.mda['coff'])
        loff = np.float32(self.mda['loff'])

        a = EQUATOR_RADIUS
        b = POLE_RADIUS
        h = ALTITUDE

        lon_0 = self.prologue['SubSatLongitude']

        nlines = int(self.mda['number_of_lines'])
        ncols = int(self.mda['number_of_columns'])

        loff = nlines - loff

        area_extent = self.get_area_extent((nlines, ncols),
                                           (loff, coff),
                                           (lfac, cfac),
                                           h)

        proj_dict = {'a': float(a),
                     'b': float(b),
                     'lon_0': float(lon_0),
                     'h': float(h),
                     'proj': 'geos',
                     'units': 'm'}

        area = geometry.AreaDefinition(
            'some_area_name',
            "On-the-fly area",
            'geosmsg',
            proj_dict,
            ncols,
            nlines,
            area_extent)

        self.area = area

        return area



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2016-2020 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Helpers for reading netcdf-based files."""

import logging

import dask.array as da
import netCDF4
import numpy as np
import xarray as xr

from satpy import CHUNK_SIZE
from satpy.readers.file_handlers import BaseFileHandler
from satpy.readers.utils import np2str

LOG = logging.getLogger(__name__)


class NetCDF4FileHandler(BaseFileHandler):
    """Small class for inspecting a NetCDF4 file and retrieving its metadata/header data.

    File information can be accessed using bracket notation. Variables are
    accessed by using:

        wrapper["var_name"]

    Or:

        wrapper["group/subgroup/var_name"]

    Attributes can be accessed by appending "/attr/attr_name" to the
    item string:

        wrapper["group/subgroup/var_name/attr/units"]

    Or for global attributes:

        wrapper["/attr/platform_short_name"]

    Or for all of global attributes:

        wrapper["/attrs"]

    Note that loading datasets requires reopening the original file
    (unless those datasets are cached, see below), but to get just the
    shape of the dataset append "/shape" to the item string:

        wrapper["group/subgroup/var_name/shape"]

    If your file has many small data variables that are frequently accessed,
    you may choose to cache some of them.  You can do this by passing a number,
    any variable smaller than this number in bytes will be read into RAM.
    Warning, this part of the API is provisional and subject to change.

    You may get an additional speedup by passing ``cache_handle=True``.  This
    will keep the netCDF4 dataset handles open throughout the lifetime of the
    object, and instead of using `xarray.open_dataset` to open every data
    variable, a dask array will be created "manually".  This may be useful if
    you have a dataset distributed over many files, such as for FCI.  Note
    that the coordinates will be missing in this case.  If you use this option,
    ``xarray_kwargs`` will have no effect.

    Args:
        filename (str): File to read
        filename_info (dict): Dictionary with filename information
        filetype_info (dict): Dictionary with filetype information
        auto_maskandscale (bool): Apply mask and scale factors
        xarray_kwargs (dict): Addition arguments to `xarray.open_dataset`
        cache_var_size (int): Cache variables smaller than this size.
        cache_handle (bool): Keep files open for lifetime of filehandler.

    """

    file_handle = None

    def __init__(self, filename, filename_info, filetype_info,
                 auto_maskandscale=False, xarray_kwargs=None,
                 cache_var_size=0, cache_handle=False):
        """Initialize object."""
        super(NetCDF4FileHandler, self).__init__(
            filename, filename_info, filetype_info)
        self.file_content = {}
        self.cached_file_content = {}
        try:
            file_handle = netCDF4.Dataset(self.filename, 'r')
        except IOError:
            LOG.exception(
                'Failed reading file %s. Possibly corrupted file', self.filename)
            raise

        self.auto_maskandscale = auto_maskandscale
        if hasattr(file_handle, "set_auto_maskandscale"):
            file_handle.set_auto_maskandscale(auto_maskandscale)

        self.collect_metadata("", file_handle)
        self.collect_dimensions("", file_handle)
        if cache_var_size > 0:
            self.collect_cache_vars(
                    [varname for (varname, var)
                        in self.file_content.items()
                        if isinstance(var, netCDF4.Variable)
                        and isinstance(var.dtype, np.dtype)  # vlen may be str
                        and var.size * var.dtype.itemsize < cache_var_size],
                    file_handle)
        if cache_handle:
            self.file_handle = file_handle
        else:
            file_handle.close()
        self._xarray_kwargs = xarray_kwargs or {}
        self._xarray_kwargs.setdefault('chunks', CHUNK_SIZE)
        self._xarray_kwargs.setdefault('mask_and_scale', self.auto_maskandscale)

    def __del__(self):
        """Delete the file handler."""
        if self.file_handle is not None:
            try:
                self.file_handle.close()
            except RuntimeError:  # presumably closed already
                pass

    def _collect_global_attrs(self, obj):
        """Collect all the global attributes for the provided file object."""
        global_attrs = {}
        for key in obj.ncattrs():
            fc_key = f"/attr/{key}"
            value = self._get_attr_value(obj, key)
            self.file_content[fc_key] = global_attrs[key] = value
        self.file_content["/attrs"] = global_attrs

    def _collect_attrs(self, name, obj):
        """Collect all the attributes for the provided file object."""
        for key in obj.ncattrs():
            fc_key = f"{name}/attr/{key}"
            value = self._get_attr_value(obj, key)
            self.file_content[fc_key] = value

    def _get_attr_value(self, obj, key):
        value = getattr(obj, key)
        try:
            value = np2str(value)
        except ValueError:
            pass
        return value

    def collect_metadata(self, name, obj):
        """Collect all file variables and attributes for the provided file object.

        This method also iterates through subgroups of the provided object.
        """
        # Look through each subgroup
        base_name = name + "/" if name else ""
        self._collect_groups_info(base_name, obj)
        self._collect_variables_info(base_name, obj)
        if not name:
            self._collect_global_attrs(obj)
        else:
            self._collect_attrs(name, obj)

    def _collect_groups_info(self, base_name, obj):
        for group_name, group_obj in obj.groups.items():
            full_group_name = base_name + group_name
            self.file_content[full_group_name] = group_obj
            self._collect_attrs(full_group_name, group_obj)
            self.collect_metadata(full_group_name, group_obj)

    def _collect_variables_info(self, base_name, obj):
        for var_name, var_obj in obj.variables.items():
            var_name = base_name + var_name
            self.file_content[var_name] = var_obj
            self.file_content[var_name + "/dtype"] = var_obj.dtype
            self.file_content[var_name + "/shape"] = var_obj.shape
            self.file_content[var_name + "/dimensions"] = var_obj.dimensions
            self._collect_attrs(var_name, var_obj)

    def collect_dimensions(self, name, obj):
        """Collect dimensions."""
        for dim_name, dim_obj in obj.dimensions.items():
            dim_name = "{}/dimension/{}".format(name, dim_name)
            self.file_content[dim_name] = len(dim_obj)

    def collect_cache_vars(self, cache_vars, obj):
        """Collect data variables for caching.

        This method will collect some data variables and store them in RAM.
        This may be useful if some small variables are frequently accessed,
        to prevent needlessly frequently opening and closing the file, which
        in case of xarray is associated with some overhead.

        Should be called later than `collect_metadata`.

        Args:
            cache_vars (List[str]): Names of data variables to be cached.
            obj (netCDF4.Dataset): Dataset object from which to read them.

        """
        for var_name in cache_vars:
            v = self.file_content[var_name]
            self.cached_file_content[var_name] = xr.DataArray(
                    v[:], dims=v.dimensions, attrs=v.__dict__, name=v.name)

    def __getitem__(self, key):
        """Get item for given key."""
        val = self.file_content[key]
        if isinstance(val, netCDF4.Variable):
            return self._get_variable(key, val)
        if isinstance(val, netCDF4.Group):
            return self._get_group(key, val)
        return val

    def _get_variable(self, key, val):
        """Get a variable from the netcdf file."""
        if key in self.cached_file_content:
            return self.cached_file_content[key]
        # these datasets are closed and inaccessible when the file is
        # closed, need to reopen
        # TODO: Handle HDF4 versus NetCDF3 versus NetCDF4
        parts = key.rsplit('/', 1)
        if len(parts) == 2:
            group, key = parts
        else:
            group = None
        if self.file_handle is not None:
            val = self._get_var_from_filehandle(group, key)
        else:
            val = self._get_var_from_xr(group, key)
        return val

    def _get_group(self, key, val):
        """Get a group from the netcdf file."""
        # Full groups are conveniently read with xr even if file_handle is available
        with xr.open_dataset(self.filename, group=key,
                             **self._xarray_kwargs) as nc:
            val = nc
        return val

    def _get_var_from_xr(self, group, key):
        with xr.open_dataset(self.filename, group=group,
                             **self._xarray_kwargs) as nc:
            val = nc[key]
            # Even though `chunks` is specified in the kwargs, xarray
            # uses dask.arrays only for data variables that have at least
            # one dimension; for zero-dimensional data variables (scalar),
            # it uses its own lazy loading for scalars.  When those are
            # accessed after file closure, xarray reopens the file without
            # closing it again.  This will leave potentially many open file
            # objects (which may in turn trigger a Segmentation Fault:
            # https://github.com/pydata/xarray/issues/2954#issuecomment-491221266
            if not val.chunks:
                val.load()
        return val

    def _get_var_from_filehandle(self, group, key):
        # Not getting coordinates as this is more work, therefore more
        # overhead, and those are not used downstream.
        if group is None:
            g = self.file_handle
        else:
            g = self.file_handle[group]
        v = g[key]
        x = xr.DataArray(
                da.from_array(v), dims=v.dimensions, attrs=v.__dict__,
                name=v.name)
        return x

    def __contains__(self, item):
        """Get item from file content."""
        return item in self.file_content

    def get(self, item, default=None):
        """Get item."""
        if item in self:
            return self[item]
        else:
            return default



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for the 'mersi2_l1b' reader."""
import os
import unittest
from unittest import mock

import dask.array as da
import numpy as np
import pytest
import xarray as xr

from satpy.tests.reader_tests.test_hdf5_utils import FakeHDF5FileHandler


class FakeHDF5FileHandler2(FakeHDF5FileHandler):
    """Swap-in HDF5 File Handler."""

    def make_test_data(self, dims):
        """Make test data."""
        return xr.DataArray(da.from_array(np.ones([dim for dim in dims], dtype=np.float32) * 10, [dim for dim in dims]))

    def _get_calibration(self, num_scans, rows_per_scan):
        calibration = {
            'Calibration/VIS_Cal_Coeff':
                xr.DataArray(
                    da.ones((19, 3), chunks=1024),
                    attrs={'Slope': np.array([1.] * 19), 'Intercept': np.array([0.] * 19)},
                    dims=('_bands', '_coeffs')),
            'Calibration/IR_Cal_Coeff':
                xr.DataArray(
                    da.ones((6, 4, num_scans), chunks=1024),
                    attrs={'Slope': np.array([1.] * 6), 'Intercept': np.array([0.] * 6)},
                    dims=('_bands', '_coeffs', '_scans')),
        }
        return calibration

    def _get_1km_data(self, num_scans, rows_per_scan, num_cols):
        data = {
            'Data/EV_1KM_RefSB':
                xr.DataArray(
                    da.ones((15, num_scans * rows_per_scan, num_cols), chunks=1024,
                            dtype=np.uint16),
                    attrs={
                        'Slope': np.array([1.] * 15), 'Intercept': np.array([0.] * 15),
                        'FillValue': 65535,
                        'units': 'NO',
                        'valid_range': [0, 4095],
                        'long_name': b'1km Earth View Science Data',
                    },
                    dims=('_ref_bands', '_rows', '_cols')),
            'Data/EV_1KM_Emissive':
                xr.DataArray(
                    da.ones((4, num_scans * rows_per_scan, num_cols), chunks=1024,
                            dtype=np.uint16),
                    attrs={
                        'Slope': np.array([1.] * 4), 'Intercept': np.array([0.] * 4),
                        'FillValue': 65535,
                        'units': 'mW/ (m2 cm-1 sr)',
                        'valid_range': [0, 25000],
                        'long_name': b'1km Emissive Bands Earth View '
                                     b'Science Data',
                    },
                    dims=('_ir_bands', '_rows', '_cols')),
            'Data/EV_250_Aggr.1KM_RefSB':
                xr.DataArray(
                    da.ones((4, num_scans * rows_per_scan, num_cols), chunks=1024,
                            dtype=np.uint16),
                    attrs={
                        'Slope': np.array([1.] * 4), 'Intercept': np.array([0.] * 4),
                        'FillValue': 65535,
                        'units': 'NO',
                        'valid_range': [0, 4095],
                        'long_name': b'250m Reflective Bands Earth View '
                                     b'Science Data Aggregated to 1 km'
                    },
                    dims=('_ref250_bands', '_rows', '_cols')),
            'Data/EV_250_Aggr.1KM_Emissive':
                xr.DataArray(
                    da.ones((2, num_scans * rows_per_scan, num_cols), chunks=1024,
                            dtype=np.uint16),
                    attrs={
                        'Slope': np.array([1.] * 2), 'Intercept': np.array([0.] * 2),
                        'FillValue': 65535,
                        'units': 'mW/ (m2 cm-1 sr)',
                        'valid_range': [0, 4095],
                        'long_name': b'250m Emissive Bands Earth View '
                                     b'Science Data Aggregated to 1 km'
                    },
                    dims=('_ir250_bands', '_rows', '_cols')),
        }
        return data

    def _get_250m_data(self, num_scans, rows_per_scan, num_cols):
        data = {
            'Data/EV_250_RefSB_b1':
                xr.DataArray(
                    da.ones((num_scans * rows_per_scan, num_cols), chunks=1024,
                            dtype=np.uint16),
                    attrs={
                        'Slope': np.array([1.] * 1), 'Intercept': np.array([0.] * 1),
                        'FillValue': 65535,
                        'units': 'NO',
                        'valid_range': [0, 4095],
                    },
                    dims=('_rows', '_cols')),
            'Data/EV_250_RefSB_b2':
                xr.DataArray(
                    da.ones((num_scans * rows_per_scan, num_cols), chunks=1024,
                            dtype=np.uint16),
                    attrs={
                        'Slope': np.array([1.] * 1), 'Intercept': np.array([0.] * 1),
                        'FillValue': 65535,
                        'units': 'NO',
                        'valid_range': [0, 4095],
                    },
                    dims=('_rows', '_cols')),
            'Data/EV_250_RefSB_b3':
                xr.DataArray(
                    da.ones((num_scans * rows_per_scan, num_cols), chunks=1024,
                            dtype=np.uint16),
                    attrs={
                        'Slope': np.array([1.] * 1), 'Intercept': np.array([0.] * 1),
                        'FillValue': 65535,
                        'units': 'NO',
                        'valid_range': [0, 4095],
                    },
                    dims=('_rows', '_cols')),
            'Data/EV_250_RefSB_b4':
                xr.DataArray(
                    da.ones((num_scans * rows_per_scan, num_cols), chunks=1024,
                            dtype=np.uint16),
                    attrs={
                        'Slope': np.array([1.] * 1), 'Intercept': np.array([0.] * 1),
                        'FillValue': 65535,
                        'units': 'NO',
                        'valid_range': [0, 4095],
                    },
                    dims=('_rows', '_cols')),
            'Data/EV_250_Emissive_b24':
                xr.DataArray(
                    da.ones((num_scans * rows_per_scan, num_cols), chunks=1024,
                            dtype=np.uint16),
                    attrs={
                        'Slope': np.array([1.] * 1), 'Intercept': np.array([0.] * 1),
                        'FillValue': 65535,
                        'units': 'mW/ (m2 cm-1 sr)',
                        'valid_range': [0, 4095],
                    },
                    dims=('_rows', '_cols')),
            'Data/EV_250_Emissive_b25':
                xr.DataArray(
                    da.ones((num_scans * rows_per_scan, num_cols), chunks=1024,
                            dtype=np.uint16),
                    attrs={
                        'Slope': np.array([1.] * 1), 'Intercept': np.array([0.] * 1),
                        'FillValue': 65535,
                        'units': 'mW/ (m2 cm-1 sr)',
                        'valid_range': [0, 4095],
                    },
                    dims=('_rows', '_cols')),
        }
        return data

    def _get_geo_data(self, num_scans, rows_per_scan, num_cols, prefix='Geolocation/'):
        geo = {
            prefix + 'Longitude':
                xr.DataArray(
                    da.ones((num_scans * rows_per_scan, num_cols), chunks=1024),
                    attrs={
                        'Slope': np.array([1.] * 1), 'Intercept': np.array([0.] * 1),
                        'units': 'degree',
                        'valid_range': [-90, 90],
                    },
                    dims=('_rows', '_cols')),
            prefix + 'Latitude':
                xr.DataArray(
                    da.ones((num_scans * rows_per_scan, num_cols), chunks=1024),
                    attrs={
                        'Slope': np.array([1.] * 1), 'Intercept': np.array([0.] * 1),
                        'units': 'degree',
                        'valid_range': [-180, 180],
                    },
                    dims=('_rows', '_cols')),
            prefix + 'SensorZenith':
                xr.DataArray(
                    da.ones((num_scans * rows_per_scan, num_cols), chunks=1024),
                    attrs={
                        'Slope': np.array([.01] * 1), 'Intercept': np.array([0.] * 1),
                        'units': 'degree',
                        'valid_range': [0, 28000],
                    },
                    dims=('_rows', '_cols')),
        }
        return geo

    def get_test_content(self, filename, filename_info, filetype_info):
        """Mimic reader input file content."""
        rows_per_scan = self.filetype_info.get('rows_per_scan', 10)
        num_scans = 2
        num_cols = 2048
        global_attrs = {
            '/attr/Observing Beginning Date': '2019-01-01',
            '/attr/Observing Ending Date': '2019-01-01',
            '/attr/Observing Beginning Time': '18:27:39.720',
            '/attr/Observing Ending Time': '18:38:36.728',
            '/attr/Satellite Name': 'FY-3D',
            '/attr/Sensor Identification Code': 'MERSI',
        }

        data = {}
        if self.filetype_info['file_type'] == 'mersi2_l1b_1000':
            data = self._get_1km_data(num_scans, rows_per_scan, num_cols)
            global_attrs['/attr/TBB_Trans_Coefficient_A'] = np.array([1.0] * 6)
            global_attrs['/attr/TBB_Trans_Coefficient_B'] = np.array([0.0] * 6)
        elif self.filetype_info['file_type'] == 'mersi2_l1b_250':
            data = self._get_250m_data(num_scans, rows_per_scan, num_cols * 2)
            global_attrs['/attr/TBB_Trans_Coefficient_A'] = np.array([0.0] * 6)
            global_attrs['/attr/TBB_Trans_Coefficient_B'] = np.array([0.0] * 6)
        elif self.filetype_info['file_type'] == 'mersi2_l1b_1000_geo':
            data = self._get_geo_data(num_scans, rows_per_scan, num_cols)
        elif self.filetype_info['file_type'] == 'mersi2_l1b_250_geo':
            data = self._get_geo_data(num_scans, rows_per_scan, num_cols * 2,
                                      prefix='')

        test_content = {}
        test_content.update(global_attrs)
        test_content.update(data)
        test_content.update(self._get_calibration(num_scans, rows_per_scan))
        return test_content


class TestMERSI2L1BReader(unittest.TestCase):
    """Test MERSI2 L1B Reader."""

    yaml_file = "mersi2_l1b.yaml"

    def setUp(self):
        """Wrap HDF5 file handler with our own fake handler."""
        from satpy._config import config_search_paths
        from satpy.readers.mersi2_l1b import MERSI2L1B
        self.reader_configs = config_search_paths(os.path.join('readers', self.yaml_file))
        # http://stackoverflow.com/questions/12219967/how-to-mock-a-base-class-with-python-mock-library
        self.p = mock.patch.object(MERSI2L1B, '__bases__', (FakeHDF5FileHandler2,))
        self.fake_handler = self.p.start()
        self.p.is_local = True

    def tearDown(self):
        """Stop wrapping the HDF5 file handler."""
        self.p.stop()

    def test_fy3d_all_resolutions(self):
        """Test loading data when all resolutions are available."""
        from satpy.dataset.data_dict import get_key
        from satpy.readers import load_reader
        from satpy.tests.utils import make_dataid
        filenames = [
            'tf2019071182739.FY3D-X_MERSI_0250M_L1B.HDF',
            'tf2019071182739.FY3D-X_MERSI_1000M_L1B.HDF',
            'tf2019071182739.FY3D-X_MERSI_GEO1K_L1B.HDF',
            'tf2019071182739.FY3D-X_MERSI_GEOQK_L1B.HDF',
        ]
        reader = load_reader(self.reader_configs)
        files = reader.select_files_from_pathnames(filenames)
        self.assertEqual(4, len(files))
        reader.create_filehandlers(files)
        # Make sure we have some files
        self.assertTrue(reader.file_handlers)

        # Verify that we have multiple resolutions for:
        #     - Bands 1-4 (visible)
        #     - Bands 24-25 (IR)
        available_datasets = reader.available_dataset_ids
        for band_name in ('1', '2', '3', '4', '24', '25'):
            if band_name in ('24', '25'):
                # don't know how to get radiance for IR bands
                num_results = 2
            else:
                num_results = 3
            ds_id = make_dataid(name=band_name, resolution=250)
            res = get_key(ds_id, available_datasets,
                          num_results=num_results, best=False)
            self.assertEqual(num_results, len(res))
            ds_id = make_dataid(name=band_name, resolution=1000)
            res = get_key(ds_id, available_datasets,
                          num_results=num_results, best=False)
            self.assertEqual(num_results, len(res))

        res = reader.load(['1', '2', '3', '4', '5', '20', '24', '25'])
        self.assertEqual(8, len(res))
        self.assertEqual((2 * 40, 2048 * 2), res['1'].shape)
        self.assertEqual('reflectance', res['1'].attrs['calibration'])
        self.assertEqual('%', res['1'].attrs['units'])
        self.assertEqual((2 * 40, 2048 * 2), res['2'].shape)
        self.assertEqual('reflectance', res['2'].attrs['calibration'])
        self.assertEqual('%', res['2'].attrs['units'])
        self.assertEqual((2 * 40, 2048 * 2), res['3'].shape)
        self.assertEqual('reflectance', res['3'].attrs['calibration'])
        self.assertEqual('%', res['3'].attrs['units'])
        self.assertEqual((2 * 40, 2048 * 2), res['4'].shape)
        self.assertEqual('reflectance', res['4'].attrs['calibration'])
        self.assertEqual('%', res['4'].attrs['units'])
        self.assertEqual((2 * 10, 2048), res['5'].shape)
        self.assertEqual('reflectance', res['5'].attrs['calibration'])
        self.assertEqual('%', res['5'].attrs['units'])
        self.assertEqual((2 * 10, 2048), res['20'].shape)
        self.assertEqual('brightness_temperature', res['20'].attrs['calibration'])
        self.assertEqual('K', res['20'].attrs['units'])
        self.assertEqual((2 * 40, 2048 * 2), res['24'].shape)
        self.assertEqual('brightness_temperature', res['24'].attrs['calibration'])
        self.assertEqual('K', res['24'].attrs['units'])
        self.assertEqual((2 * 40, 2048 * 2), res['25'].shape)
        self.assertEqual('brightness_temperature', res['25'].attrs['calibration'])
        self.assertEqual('K', res['25'].attrs['units'])

    def test_fy3d_counts_calib(self):
        """Test loading data at counts calibration."""
        from satpy.readers import load_reader
        from satpy.tests.utils import make_dataid
        filenames = [
            'tf2019071182739.FY3D-X_MERSI_0250M_L1B.HDF',
            'tf2019071182739.FY3D-X_MERSI_1000M_L1B.HDF',
            'tf2019071182739.FY3D-X_MERSI_GEO1K_L1B.HDF',
            'tf2019071182739.FY3D-X_MERSI_GEOQK_L1B.HDF',
        ]
        reader = load_reader(self.reader_configs)
        files = reader.select_files_from_pathnames(filenames)
        self.assertEqual(4, len(files))
        reader.create_filehandlers(files)
        # Make sure we have some files
        self.assertTrue(reader.file_handlers)

        ds_ids = []
        for band_name in ['1', '2', '3', '4', '5', '20', '24', '25']:
            ds_ids.append(make_dataid(name=band_name, calibration='counts'))
        ds_ids.append(make_dataid(name='satellite_zenith_angle'))
        res = reader.load(ds_ids)
        self.assertEqual(9, len(res))
        self.assertEqual((2 * 40, 2048 * 2), res['1'].shape)
        self.assertEqual('counts', res['1'].attrs['calibration'])
        self.assertEqual(res['1'].dtype, np.uint16)
        self.assertEqual('1', res['1'].attrs['units'])
        self.assertEqual((2 * 40, 2048 * 2), res['2'].shape)
        self.assertEqual('counts', res['2'].attrs['calibration'])
        self.assertEqual(res['2'].dtype, np.uint16)
        self.assertEqual('1', res['2'].attrs['units'])
        self.assertEqual((2 * 40, 2048 * 2), res['3'].shape)
        self.assertEqual('counts', res['3'].attrs['calibration'])
        self.assertEqual(res['3'].dtype, np.uint16)
        self.assertEqual('1', res['3'].attrs['units'])
        self.assertEqual((2 * 40, 2048 * 2), res['4'].shape)
        self.assertEqual('counts', res['4'].attrs['calibration'])
        self.assertEqual(res['4'].dtype, np.uint16)
        self.assertEqual('1', res['4'].attrs['units'])
        self.assertEqual((2 * 10, 2048), res['5'].shape)
        self.assertEqual('counts', res['5'].attrs['calibration'])
        self.assertEqual(res['5'].dtype, np.uint16)
        self.assertEqual('1', res['5'].attrs['units'])
        self.assertEqual((2 * 10, 2048), res['20'].shape)
        self.assertEqual('counts', res['20'].attrs['calibration'])
        self.assertEqual(res['20'].dtype, np.uint16)
        self.assertEqual('1', res['20'].attrs['units'])
        self.assertEqual((2 * 40, 2048 * 2), res['24'].shape)
        self.assertEqual('counts', res['24'].attrs['calibration'])
        self.assertEqual(res['24'].dtype, np.uint16)
        self.assertEqual('1', res['24'].attrs['units'])
        self.assertEqual((2 * 40, 2048 * 2), res['25'].shape)
        self.assertEqual('counts', res['25'].attrs['calibration'])
        self.assertEqual(res['25'].dtype, np.uint16)
        self.assertEqual('1', res['25'].attrs['units'])

    def test_fy3d_rad_calib(self):
        """Test loading data at radiance calibration."""
        from satpy.readers import load_reader
        from satpy.tests.utils import make_dataid
        filenames = [
            'tf2019071182739.FY3D-X_MERSI_0250M_L1B.HDF',
            'tf2019071182739.FY3D-X_MERSI_1000M_L1B.HDF',
            'tf2019071182739.FY3D-X_MERSI_GEO1K_L1B.HDF',
            'tf2019071182739.FY3D-X_MERSI_GEOQK_L1B.HDF',
        ]
        reader = load_reader(self.reader_configs)
        files = reader.select_files_from_pathnames(filenames)
        self.assertEqual(4, len(files))
        reader.create_filehandlers(files)
        # Make sure we have some files
        self.assertTrue(reader.file_handlers)

        ds_ids = []
        for band_name in ['1', '2', '3', '4', '5']:
            ds_ids.append(make_dataid(name=band_name, calibration='radiance'))
        res = reader.load(ds_ids)
        self.assertEqual(5, len(res))
        self.assertEqual((2 * 40, 2048 * 2), res['1'].shape)
        self.assertEqual('radiance', res['1'].attrs['calibration'])
        self.assertEqual('mW/ (m2 cm-1 sr)', res['1'].attrs['units'])
        self.assertEqual((2 * 40, 2048 * 2), res['2'].shape)
        self.assertEqual('radiance', res['2'].attrs['calibration'])
        self.assertEqual('mW/ (m2 cm-1 sr)', res['2'].attrs['units'])
        self.assertEqual((2 * 40, 2048 * 2), res['3'].shape)
        self.assertEqual('radiance', res['3'].attrs['calibration'])
        self.assertEqual('mW/ (m2 cm-1 sr)', res['3'].attrs['units'])
        self.assertEqual((2 * 40, 2048 * 2), res['4'].shape)
        self.assertEqual('radiance', res['4'].attrs['calibration'])
        self.assertEqual('mW/ (m2 cm-1 sr)', res['4'].attrs['units'])
        self.assertEqual((2 * 10, 2048), res['5'].shape)
        self.assertEqual('radiance', res['5'].attrs['calibration'])
        self.assertEqual('mW/ (m2 cm-1 sr)', res['5'].attrs['units'])

    def test_fy3d_1km_resolutions(self):
        """Test loading data when only 1km resolutions are available."""
        from satpy.dataset.data_dict import get_key
        from satpy.readers import load_reader
        from satpy.tests.utils import make_dataid
        filenames = [
            'tf2019071182739.FY3D-X_MERSI_1000M_L1B.HDF',
            'tf2019071182739.FY3D-X_MERSI_GEO1K_L1B.HDF',
        ]
        reader = load_reader(self.reader_configs)
        files = reader.select_files_from_pathnames(filenames)
        self.assertEqual(2, len(files))
        reader.create_filehandlers(files)
        # Make sure we have some files
        self.assertTrue(reader.file_handlers)

        # Verify that we have multiple resolutions for:
        #     - Bands 1-4 (visible)
        #     - Bands 24-25 (IR)
        available_datasets = reader.available_dataset_ids
        for band_name in ('1', '2', '3', '4', '24', '25'):
            if band_name in ('24', '25'):
                # don't know how to get radiance for IR bands
                num_results = 2
            else:
                num_results = 3
            ds_id = make_dataid(name=band_name, resolution=250)
            with pytest.raises(KeyError):
                res = get_key(ds_id, available_datasets,
                              num_results=num_results, best=False)
            ds_id = make_dataid(name=band_name, resolution=1000)
            res = get_key(ds_id, available_datasets,
                          num_results=num_results, best=False)
            self.assertEqual(num_results, len(res))

        res = reader.load(['1', '2', '3', '4', '5', '20', '24', '25'])
        self.assertEqual(8, len(res))
        self.assertEqual((2 * 10, 2048), res['1'].shape)
        self.assertEqual('reflectance', res['1'].attrs['calibration'])
        self.assertEqual('%', res['1'].attrs['units'])
        self.assertEqual((2 * 10, 2048), res['2'].shape)
        self.assertEqual('reflectance', res['2'].attrs['calibration'])
        self.assertEqual('%', res['2'].attrs['units'])
        self.assertEqual((2 * 10, 2048), res['3'].shape)
        self.assertEqual('reflectance', res['3'].attrs['calibration'])
        self.assertEqual('%', res['3'].attrs['units'])
        self.assertEqual((2 * 10, 2048), res['4'].shape)
        self.assertEqual('reflectance', res['4'].attrs['calibration'])
        self.assertEqual('%', res['4'].attrs['units'])
        self.assertEqual((2 * 10, 2048), res['5'].shape)
        self.assertEqual('reflectance', res['5'].attrs['calibration'])
        self.assertEqual('%', res['5'].attrs['units'])
        self.assertEqual((2 * 10, 2048), res['20'].shape)
        self.assertEqual('brightness_temperature', res['20'].attrs['calibration'])
        self.assertEqual('K', res['20'].attrs['units'])
        self.assertEqual((2 * 10, 2048), res['24'].shape)
        self.assertEqual('brightness_temperature', res['24'].attrs['calibration'])
        self.assertEqual('K', res['24'].attrs['units'])
        self.assertEqual((2 * 10, 2048), res['25'].shape)
        self.assertEqual('brightness_temperature', res['25'].attrs['calibration'])
        self.assertEqual('K', res['25'].attrs['units'])

    def test_fy3d_250_resolutions(self):
        """Test loading data when only 250m resolutions are available."""
        from satpy.dataset.data_dict import get_key
        from satpy.readers import load_reader
        from satpy.tests.utils import make_dataid
        filenames = [
            'tf2019071182739.FY3D-X_MERSI_0250M_L1B.HDF',
            'tf2019071182739.FY3D-X_MERSI_GEOQK_L1B.HDF',
        ]
        reader = load_reader(self.reader_configs)
        files = reader.select_files_from_pathnames(filenames)
        self.assertEqual(2, len(files))
        reader.create_filehandlers(files)
        # Make sure we have some files
        self.assertTrue(reader.file_handlers)

        # Verify that we have multiple resolutions for:
        #     - Bands 1-4 (visible)
        #     - Bands 24-25 (IR)
        available_datasets = reader.available_dataset_ids
        for band_name in ('1', '2', '3', '4', '24', '25'):
            if band_name in ('24', '25'):
                # don't know how to get radiance for IR bands
                num_results = 2
            else:
                num_results = 3
            ds_id = make_dataid(name=band_name, resolution=250)
            res = get_key(ds_id, available_datasets,
                          num_results=num_results, best=False)
            self.assertEqual(num_results, len(res))
            ds_id = make_dataid(name=band_name, resolution=1000)
            with pytest.raises(KeyError):
                res = get_key(ds_id, available_datasets,
                              num_results=num_results, best=False)

        res = reader.load(['1', '2', '3', '4', '5', '20', '24', '25'])
        self.assertEqual(6, len(res))
        self.assertRaises(KeyError, res.__getitem__, '5')
        self.assertRaises(KeyError, res.__getitem__, '20')
        self.assertEqual((2 * 40, 2048 * 2), res['1'].shape)
        self.assertEqual('reflectance', res['1'].attrs['calibration'])
        self.assertEqual('%', res['1'].attrs['units'])
        self.assertEqual((2 * 40, 2048 * 2), res['2'].shape)
        self.assertEqual('reflectance', res['2'].attrs['calibration'])
        self.assertEqual('%', res['2'].attrs['units'])
        self.assertEqual((2 * 40, 2048 * 2), res['3'].shape)
        self.assertEqual('reflectance', res['3'].attrs['calibration'])
        self.assertEqual('%', res['3'].attrs['units'])
        self.assertEqual((2 * 40, 2048 * 2), res['4'].shape)
        self.assertEqual('reflectance', res['4'].attrs['calibration'])
        self.assertEqual('%', res['4'].attrs['units'])
        self.assertEqual((2 * 40, 2048 * 2), res['24'].shape)
        self.assertEqual('brightness_temperature', res['24'].attrs['calibration'])
        self.assertEqual('K', res['24'].attrs['units'])
        self.assertEqual((2 * 40, 2048 * 2), res['25'].shape)
        self.assertEqual('brightness_temperature', res['25'].attrs['calibration'])
        self.assertEqual('K', res['25'].attrs['units'])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2011-2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Interface to VIIRS L1B format."""

import logging
from datetime import datetime

import numpy as np

from satpy.readers.netcdf_utils import NetCDF4FileHandler

LOG = logging.getLogger(__name__)


class VIIRSL1BFileHandler(NetCDF4FileHandler):
    """VIIRS L1B File Reader."""

    def _parse_datetime(self, datestr):
        """Parse datetime."""
        return datetime.strptime(datestr, "%Y-%m-%dT%H:%M:%S.000Z")

    @property
    def start_orbit_number(self):
        """Get start orbit number."""
        try:
            return int(self['/attr/orbit_number'])
        except KeyError:
            return int(self['/attr/OrbitNumber'])

    @property
    def end_orbit_number(self):
        """Get end orbit number."""
        try:
            return int(self['/attr/orbit_number'])
        except KeyError:
            return int(self['/attr/OrbitNumber'])

    @property
    def platform_name(self):
        """Get platform name."""
        try:
            res = self.get('/attr/platform',
                           self.filename_info['platform_shortname'])
        except KeyError:
            res = 'Unknown'

        return {
            'JPSS-1': 'NOAA-20',
            'NP': 'Suomi-NPP',
            'J1': 'NOAA-20',
            'J2': 'NOAA-21',
            'JPSS-2': 'NOAA-21',
        }.get(res, res)

    @property
    def sensor_name(self):
        """Get sensor name."""
        res = self['/attr/instrument']
        if isinstance(res, np.ndarray):
            return str(res.astype(str))
        else:
            return res

    def adjust_scaling_factors(self, factors, file_units, output_units):
        """Adjust scaling factors."""
        if factors is None or factors[0] is None:
            factors = [1, 0]
        if file_units == output_units:
            LOG.debug("File units and output units are the same (%s)", file_units)
            return factors
        factors = np.array(factors)

        if file_units == "W cm-2 sr-1" and output_units == "W m-2 sr-1":
            LOG.debug("Adjusting scaling factors to convert '%s' to '%s'", file_units, output_units)
            factors[::2] = np.where(factors[::2] != -999, factors[::2] * 10000.0, -999)
            factors[1::2] = np.where(factors[1::2] != -999, factors[1::2] * 10000.0, -999)
            return factors
        elif file_units == "1" and output_units == "%":
            LOG.debug("Adjusting scaling factors to convert '%s' to '%s'", file_units, output_units)
            factors[::2] = np.where(factors[::2] != -999, factors[::2] * 100.0, -999)
            factors[1::2] = np.where(factors[1::2] != -999, factors[1::2] * 100.0, -999)
            return factors
        else:
            return factors

    def get_shape(self, ds_id, ds_info):
        """Get shape."""
        var_path = ds_info.get('file_key', 'observation_data/{}'.format(ds_id['name']))
        return self.get(var_path + '/shape', 1)

    @property
    def start_time(self):
        """Get start time."""
        return self._parse_datetime(self['/attr/time_coverage_start'])

    @property
    def end_time(self):
        """Get end time."""
        return self._parse_datetime(self['/attr/time_coverage_end'])

    def _get_dataset_file_units(self, dataset_id, ds_info, var_path):
        file_units = ds_info.get('file_units')
        if file_units is None:
            file_units = self.get(var_path + '/attr/units')
            # they were almost completely CF compliant...
            if file_units == "none":
                file_units = "1"

        if dataset_id.get('calibration') == 'radiance' and ds_info['units'] == 'W m-2 um-1 sr-1':
            rad_units_path = var_path + '/attr/radiance_units'
            if rad_units_path in self:
                if file_units is None:
                    file_units = self[var_path + '/attr/radiance_units']
                if file_units == 'Watts/meter^2/steradian/micrometer':
                    file_units = 'W m-2 um-1 sr-1'
        elif ds_info.get('units') == '%' and file_units is None:
            # v1.1 and above of level 1 processing removed 'units' attribute
            # for all reflectance channels
            file_units = "1"

        return file_units

    def _get_dataset_valid_range(self, dataset_id, ds_info, var_path):
        if dataset_id.get('calibration') == 'radiance' and ds_info['units'] == 'W m-2 um-1 sr-1':
            rad_units_path = var_path + '/attr/radiance_units'
            if rad_units_path in self:
                # we are getting a reflectance band but we want the radiance values
                # special scaling parameters
                scale_factor = self[var_path + '/attr/radiance_scale_factor']
                scale_offset = self[var_path + '/attr/radiance_add_offset']
            else:
                # we are getting a btemp band but we want the radiance values
                # these are stored directly in the primary variable
                scale_factor = self[var_path + '/attr/scale_factor']
                scale_offset = self[var_path + '/attr/add_offset']
            valid_min = self[var_path + '/attr/valid_min']
            valid_max = self[var_path + '/attr/valid_max']
        elif ds_info.get('units') == '%':
            # normal reflectance
            valid_min = self[var_path + '/attr/valid_min']
            valid_max = self[var_path + '/attr/valid_max']
            scale_factor = self[var_path + '/attr/scale_factor']
            scale_offset = self[var_path + '/attr/add_offset']
        elif ds_info.get('units') == 'K':
            # normal brightness temperature
            # use a special LUT to get the actual values
            lut_var_path = ds_info.get('lut', var_path + '_brightness_temperature_lut')
            # we get the BT values from a look up table using the scaled radiance integers
            valid_min = self[lut_var_path + '/attr/valid_min']
            valid_max = self[lut_var_path + '/attr/valid_max']
            scale_factor = scale_offset = None
        else:
            valid_min = self.get(var_path + '/attr/valid_min')
            valid_max = self.get(var_path + '/attr/valid_max')
            scale_factor = self.get(var_path + '/attr/scale_factor')
            scale_offset = self.get(var_path + '/attr/add_offset')

        return valid_min, valid_max, scale_factor, scale_offset

    def get_metadata(self, dataset_id, ds_info):
        """Get metadata."""
        var_path = ds_info.get('file_key', 'observation_data/{}'.format(dataset_id['name']))
        shape = self.get_shape(dataset_id, ds_info)
        file_units = self._get_dataset_file_units(dataset_id, ds_info, var_path)

        # Get extra metadata
        if '/dimension/number_of_scans' in self:
            rows_per_scan = int(shape[0] / self['/dimension/number_of_scans'])
            ds_info.setdefault('rows_per_scan', rows_per_scan)

        i = getattr(self[var_path], 'attrs', {})
        i.update(ds_info)
        i.update(dataset_id.to_dict())
        i.update({
            "shape": shape,
            "units": ds_info.get("units", file_units),
            "file_units": file_units,
            "platform_name": self.platform_name,
            "sensor": self.sensor_name,
            "start_orbit": self.start_orbit_number,
            "end_orbit": self.end_orbit_number,
        })
        i.update(dataset_id.to_dict())
        return i

    def get_dataset(self, dataset_id, ds_info):
        """Get dataset."""
        var_path = ds_info.get('file_key', 'observation_data/{}'.format(dataset_id['name']))
        metadata = self.get_metadata(dataset_id, ds_info)

        valid_min, valid_max, scale_factor, scale_offset = self._get_dataset_valid_range(dataset_id, ds_info, var_path)
        if dataset_id.get('calibration') == 'radiance' and ds_info['units'] == 'W m-2 um-1 sr-1':
            data = self[var_path]
        elif ds_info.get('units') == '%':
            data = self[var_path]
        elif ds_info.get('units') == 'K':
            # normal brightness temperature
            # use a special LUT to get the actual values
            lut_var_path = ds_info.get('lut', var_path + '_brightness_temperature_lut')
            data = self[var_path]
            # we get the BT values from a look up table using the scaled radiance integers
            index_arr = data.data.astype(int)
            coords = data.coords
            data.data = self[lut_var_path].data[index_arr.ravel()].reshape(data.shape)
            data = data.assign_coords(**coords)
        else:
            data = self[var_path]
        data.attrs.update(metadata)

        if valid_min is not None and valid_max is not None:
            data = data.where((data >= valid_min) & (data <= valid_max))
        if data.attrs.get('units') in ['%', 'K', '1', 'W m-2 um-1 sr-1'] and \
                'flag_meanings' in data.attrs:
            # flag meanings don't mean anything anymore for these variables
            # these aren't category products
            data.attrs.pop('flag_meanings', None)
            data.attrs.pop('flag_values', None)

        factors = (scale_factor, scale_offset)
        factors = self.adjust_scaling_factors(factors, metadata['file_units'], ds_info.get("units"))
        if factors[0] != 1 or factors[1] != 0:
            data *= factors[0]
            data += factors[1]
        # rename dimensions to correspond to satpy's 'y' and 'x' standard
        if 'number_of_lines' in data.dims:
            data = data.rename({'number_of_lines': 'y', 'number_of_pixels': 'x'})
        return data



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2020, 2021 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Module for testing the satpy.readers.hy2_scat_l2b_h5 module."""

import os
import unittest
from unittest import mock

import dask.array as da
import numpy as np
import xarray as xr

from satpy.tests.reader_tests.test_hdf5_utils import FakeHDF5FileHandler

DEFAULT_FILE_DTYPE = np.uint16
DEFAULT_FILE_SHAPE = (10, 300)
DEFAULT_LAT_DATA = np.linspace(45, 65, DEFAULT_FILE_SHAPE[1]).astype(np.float32)
DEFAULT_LAT_DATA = np.repeat([DEFAULT_LAT_DATA], DEFAULT_FILE_SHAPE[0], axis=0)
DEFAULT_LON_DATA = np.linspace(-10, 10, DEFAULT_FILE_SHAPE[1]).astype(np.float32)
DEFAULT_LON_DATA = np.repeat([DEFAULT_LON_DATA], DEFAULT_FILE_SHAPE[0], axis=0)
DEFAULT_FILE_DATA = np.arange(DEFAULT_FILE_SHAPE[0] * DEFAULT_FILE_SHAPE[1],
                              dtype=DEFAULT_FILE_DTYPE).reshape(DEFAULT_FILE_SHAPE)


class FakeHDF5FileHandler2(FakeHDF5FileHandler):
    """Swap-in HDF5 File Handler."""

    def _get_geo_data(self, num_rows, num_cols):
        geo = {
            'wvc_lon':
                xr.DataArray(
                    da.ones((num_rows, num_cols), chunks=1024,
                            dtype=np.float32),
                    attrs={
                        'fill_value': 1.7e+38,
                        'scale_factor': 1.,
                        'add_offset': 0.,
                        'units': 'degree',
                        'valid range': [0, 359.99],
                    },
                    dims=('y', 'x')),
            'wvc_lat':
                xr.DataArray(
                    da.ones((num_rows, num_cols), chunks=1024,
                            dtype=np.float32),
                    attrs={
                        'fill_value': 1.7e+38,
                        'scale_factor': 1.,
                        'add_offset': 0.,
                        'units': 'degree',
                        'valid range': [-90.0, 90.0],
                    },
                    dims=('y', 'x')),
        }
        return geo

    def _get_geo_data_nsoas(self, num_rows, num_cols):
        geo = {
            'wvc_lon':
                xr.DataArray(
                    da.ones((num_rows, num_cols), chunks=1024,
                            dtype=np.float32),
                    attrs={
                        'fill_value': 1.7e+38,
                        'scale_factor': 1.,
                        'add_offset': 0.,
                        'units': 'degree',
                        'valid_range': [0, 359.99],
                    },
                    dims=('y', 'x')),
            'wvc_lat':
                xr.DataArray(
                    da.ones((num_rows, num_cols), chunks=1024,
                            dtype=np.float32),
                    attrs={
                        'fill_value': 1.7e+38,
                        'scale_factor': 1.,
                        'add_offset': 0.,
                        'units': 'degree',
                        'valid_range': [-90.0, 90.0],
                    },
                    dims=('y', 'x')),
        }
        return geo

    def _get_selection_data(self, num_rows, num_cols):
        selection = {
            'wvc_selection':
            xr.DataArray(
                da.ones((num_rows, num_cols), chunks=1024,
                        dtype=np.int8),
                attrs={
                    'fill_value': 0,
                    'scale_factor': 1.,
                    'add_offset': 0.,
                    'units': 'count',
                    'valid range': [1, 8],
                },
                dims=('y', 'x')),
            'wind_speed_selection':
            xr.DataArray(
                da.ones((num_rows, num_cols), chunks=1024,
                        dtype=np.int16),
                attrs={
                    'fill_value': -32767,
                    'scale_factor': 0.1,
                    'add_offset': 0.,
                    'units': 'deg',
                    'valid range': [0, 3599],
                },
                dims=('y', 'x')),
            'wind_dir_selection':
            xr.DataArray(
                da.ones((num_rows, num_cols), chunks=1024,
                        dtype=np.int16),
                attrs={
                    'fill_value': -32767,
                    'scale_factor': 0.01,
                    'add_offset': 0.,
                    'units': 'm/s',
                    'valid range': [0, 5000],
                },
                dims=('y', 'x')),
            'model_dir':
            xr.DataArray(
                da.ones((num_rows, num_cols), chunks=1024,
                        dtype=np.int16),
                attrs={
                    'fill_value': -32767,
                    'scale_factor': 0.01,
                    'add_offset': 0.,
                    'units': 'm/s',
                    'valid range': [0, 5000],
                },
                dims=('y', 'x')),
            'model_speed':
            xr.DataArray(
                da.ones((num_rows, num_cols), chunks=1024,
                        dtype=np.int16),
                attrs={
                    'fill_value': -32767,
                    'scale_factor': 0.1,
                    'add_offset': 0.,
                    'units': 'deg',
                    'valid range': [0, 3599],
                },
                dims=('y', 'x')),
            'num_ambigs':
            xr.DataArray(
                da.ones((num_rows, num_cols), chunks=1024,
                        dtype=np.int8),
                attrs={
                    'fill_value': 0,
                    'scale_factor': 1.,
                    'add_offset': 0.,
                    'units': 'count',
                    'valid range': [1, 8],
                },
                dims=('y', 'x')),
            'num_in_aft':
            xr.DataArray(
                da.ones((num_rows, num_cols), chunks=1024,
                        dtype=np.int8),
                attrs={
                    'fill_value': 0,
                    'scale_factor': 1.,
                    'add_offset': 0.,
                    'units': 'count',
                    'valid range': [1, 127],
                },
                dims=('y', 'x')),
            'num_in_fore':
            xr.DataArray(
                da.ones((num_rows, num_cols), chunks=1024,
                        dtype=np.int8),
                attrs={
                    'fill_value': 0,
                    'scale_factor': 1.,
                    'add_offset': 0.,
                    'units': 'count',
                    'valid range': [1, 127],
                },
                dims=('y', 'x')),
            'num_out_aft':
            xr.DataArray(
                da.ones((num_rows, num_cols), chunks=1024,
                        dtype=np.int8),
                attrs={
                    'fill_value': 0,
                    'scale_factor': 1.,
                    'add_offset': 0.,
                    'units': 'count',
                    'valid range': [1, 127],
                },
                dims=('y', 'x')),
            'num_out_fore':
            xr.DataArray(
                da.ones((num_rows, num_cols), chunks=1024,
                        dtype=np.int8),
                attrs={
                    'fill_value': 0,
                    'scale_factor': 1.,
                    'add_offset': 0.,
                    'units': 'count',
                    'valid range': [1, 127],
                },
                dims=('y', 'x')),
            'wvc_quality_flag':
            xr.DataArray(
                da.ones((num_rows, num_cols), chunks=1024,
                        dtype=np.uint16),
                attrs={
                    'fill_value': 2.14748e+09,
                    'scale_factor': 1.,
                    'add_offset': 0.,
                    'units': 'na',
                    'valid range': [1, 2.14748e+09],
                },
                dims=('y', 'x')),
        }
        return selection

    def _get_all_ambiguities_data(self, num_rows, num_cols, num_amb):
        all_amb = {
            'max_likelihood_est':
                xr.DataArray(
                    da.ones((num_rows, num_cols, num_amb), chunks=1024,
                            dtype=np.int16),
                    attrs={
                        'fill_value': -32767,
                        'scale_factor': 1.,
                        'add_offset': 0.,
                        'units': 'na',
                        'valid range': [0, 32767],
                    },
                    dims=('y', 'x', 'selection')),
            'wind_dir':
                xr.DataArray(
                    da.ones((num_rows, num_cols, num_amb), chunks=1024,
                            dtype=np.int16),
                    attrs={
                        'fill_value': -32767,
                        'scale_factor': 0.1,
                        'add_offset': 0.,
                        'units': 'deg',
                        'valid range': [0, 3599],
                    },
                    dims=('y', 'x', 'selection')),
            'wind_speed':
                xr.DataArray(
                    da.ones((num_rows, num_cols, num_amb), chunks=1024,
                            dtype=np.int16),
                    attrs={
                        'fill_value': -32767,
                        'scale_factor': 0.01,
                        'add_offset': 0.,
                        'units': 'm/s',
                        'valid range': [0, 5000],
                    },
                    dims=('y', 'x', 'selection')),
        }
        return all_amb

    def _get_wvc_row_time(self, num_rows):
        data = ["20200326T01:11:07.639",
                "20200326T01:11:11.443",
                "20200326T01:11:15.246",
                "20200326T01:11:19.049",
                "20200326T01:11:22.856",
                "20200326T01:11:26.660",
                "20200326T01:11:30.464",
                "20200326T01:11:34.268",
                "20200326T01:11:38.074",
                "20200326T01:11:41.887"]
        wvc_row_time = {
            'wvc_row_time':
                xr.DataArray(data,
                             attrs={
                                 'fill_value': "",
                             },
                             dims=('y',)),
        }
        return wvc_row_time

    def _get_global_attrs(self, num_rows, num_cols):
        return {
            '/attr/Equator_Crossing_Longitude': '246.408397',
            '/attr/Equator_Crossing_Time': '20200326T01:37:15.875',
            '/attr/HDF_Version_Id': 'HDF5-1.8.16',
            '/attr/Input_L2A_Filename': 'H2B_OPER_SCA_L2A_OR_20200326T010839_20200326T025757_07076_dps_250_20.h5',
            '/attr/Instrument_ShorName': 'HSCAT-B',
            '/attr/L2A_Inputdata_Version': '10',
            '/attr/L2B_Actual_WVC_Rows': np.int32(num_rows),
            '/attr/L2B_Algorithm_Descriptor': ('Wind retrieval processing uses the multiple solution scheme (MSS) for '
                                               'wind inversion with the NSCAT-4 GMF,and a circular median filter '
                                               'method (CMF) for ambiguity removal. The ECMWF/NCEP forescate data are '
                                               'used as background winds in the CMF'),
            '/attr/L2B_Data_Version': '10',
            '/attr/L2B_Expected_WVC_Rows': np.int32(num_rows),
            '/attr/L2B_Processing_Type': 'OPER',
            '/attr/L2B_Processor_Name': 'hy2_sca_l2b_pro',
            '/attr/L2B_Processor_Version': '01.00',
            '/attr/Long_Name': 'HY-2B/SCAT Level 2B Ocean Wind Vectors in 25.0 km Swath Grid',
            '/attr/Orbit_Inclination': np.float32(99.3401),
            '/attr/Orbit_Number': '07076',
            '/attr/Output_L2B_Filename': 'H2B_OPER_SCA_L2B_OR_20200326T011107_20200326T025540_07076_dps_250_20_owv.h5',
            '/attr/Platform_LongName': 'Haiyang 2B Ocean Observing Satellite',
            '/attr/Platform_ShortName': 'HY-2B',
            '/attr/Platform_Type': 'spacecraft',
            '/attr/Producer_Agency': 'Ministry of Natural Resources of the People\'s Republic of China',
            '/attr/Producer_Institution': 'NSOAS',
            '/attr/Production_Date_Time': '20200326T06:23:10',
            '/attr/Range_Beginning_Time': '20200326T01:11:07',
            '/attr/Range_Ending_Time': '20200326T02:55:40',
            '/attr/Rev_Orbit_Period': '14 days',
            '/attr/Short_Name': 'HY-2B SCAT-L2B-25km',
            '/attr/Sigma0_Granularity': 'whole pulse',
            '/attr/WVC_Size': '25000m*25000m',
        }

    def get_test_content(self, filename, filename_info, filetype_info):
        """Mimic reader input file content."""
        num_rows = 300
        num_cols = 10
        num_amb = 8

        test_content = {}
        test_content.update(self._get_global_attrs(num_rows, num_cols))
        data = {}
        if 'OPER_SCA_L2B' in filename:
            test_content.update({'/attr/L2B_Expected_WVC_Cells': np.int32(num_cols)})
            data = self._get_geo_data_nsoas(num_rows, num_cols)
        else:
            test_content.update({'/attr/L2B_Number_WVC_cells': np.int32(num_cols)})
            data = self._get_geo_data(num_rows, num_cols)

        test_content.update(data)
        data = self._get_selection_data(num_rows, num_cols)
        test_content.update(data)
        data = self._get_all_ambiguities_data(num_rows, num_cols, num_amb)
        test_content.update(data)

        data = self._get_wvc_row_time(num_rows)
        test_content.update(data)

        return test_content


class TestHY2SCATL2BH5Reader(unittest.TestCase):
    """Test HY2 Scatterometer L2B H5 Reader."""

    yaml_file = "hy2_scat_l2b_h5.yaml"

    def setUp(self):
        """Wrap HDF5 file handler with our own fake handler."""
        from satpy._config import config_search_paths
        from satpy.readers.hy2_scat_l2b_h5 import HY2SCATL2BH5FileHandler
        self.reader_configs = config_search_paths(os.path.join('readers', self.yaml_file))
        # http://stackoverflow.com/questions/12219967/how-to-mock-a-base-class-with-python-mock-library
        self.p = mock.patch.object(HY2SCATL2BH5FileHandler, '__bases__', (FakeHDF5FileHandler2,))
        self.fake_handler = self.p.start()
        self.p.is_local = True

    def tearDown(self):
        """Stop wrapping the HDF5 file handler."""
        self.p.stop()

    def test_load_geo(self):
        """Test loading data."""
        from satpy.readers import load_reader
        filenames = [
            'W_XX-EUMETSAT-Darmstadt,SURFACE+SATELLITE,HY2B+SM_C_EUMP_20200326------_07077_o_250_l2b.h5', ]

        reader = load_reader(self.reader_configs)
        files = reader.select_files_from_pathnames(filenames)
        self.assertEqual(1, len(files))
        reader.create_filehandlers(files)
        # Make sure we have some files
        self.assertTrue(reader.file_handlers)

        res = reader.load(['wvc_lon', 'wvc_lat'])
        self.assertEqual(2, len(res))

    def test_load_geo_nsoas(self):
        """Test loading data from nsoas file."""
        from satpy.readers import load_reader
        filenames = [
            'H2B_OPER_SCA_L2B_OR_20210803T100304_20210803T104601_13905_pwp_250_07_owv.h5', ]

        reader = load_reader(self.reader_configs)
        files = reader.select_files_from_pathnames(filenames)
        self.assertEqual(1, len(files))
        reader.create_filehandlers(files)
        # Make sure we have some files
        self.assertTrue(reader.file_handlers)

        res = reader.load(['wvc_lon', 'wvc_lat'])
        self.assertEqual(2, len(res))

    def test_load_data_selection(self):
        """Test loading data."""
        from satpy.readers import load_reader
        filenames = [
            'W_XX-EUMETSAT-Darmstadt,SURFACE+SATELLITE,HY2B+SM_C_EUMP_20200326------_07077_o_250_l2b.h5', ]

        reader = load_reader(self.reader_configs)
        files = reader.select_files_from_pathnames(filenames)
        self.assertEqual(1, len(files))
        reader.create_filehandlers(files)
        # Make sure we have some files
        self.assertTrue(reader.file_handlers)
        res = reader.load(['wind_speed_selection',
                           'wind_dir_selection',
                           'wvc_selection'])
        self.assertEqual(3, len(res))

    def test_load_data_all_ambiguities(self):
        """Test loading data."""
        from satpy.readers import load_reader
        filenames = [
            'W_XX-EUMETSAT-Darmstadt,SURFACE+SATELLITE,HY2B+SM_C_EUMP_20200326------_07077_o_250_l2b.h5', ]

        reader = load_reader(self.reader_configs)
        files = reader.select_files_from_pathnames(filenames)
        self.assertEqual(1, len(files))
        reader.create_filehandlers(files)
        # Make sure we have some files
        self.assertTrue(reader.file_handlers)
        res = reader.load(['wind_speed',
                           'wind_dir',
                           'max_likelihood_est',
                           'model_dir',
                           'model_speed',
                           'num_ambigs',
                           'num_in_aft',
                           'num_in_fore',
                           'num_out_aft',
                           'num_out_fore',
                           'wvc_quality_flag'])
        self.assertEqual(11, len(res))

    def test_load_data_row_times(self):
        """Test loading data."""
        from satpy.readers import load_reader
        filenames = [
            'W_XX-EUMETSAT-Darmstadt,SURFACE+SATELLITE,HY2B+SM_C_EUMP_20200326------_07077_o_250_l2b.h5', ]

        reader = load_reader(self.reader_configs)
        files = reader.select_files_from_pathnames(filenames)
        self.assertEqual(1, len(files))
        reader.create_filehandlers(files)
        # Make sure we have some files
        self.assertTrue(reader.file_handlers)
        res = reader.load(['wvc_row_time'])
        self.assertEqual(1, len(res))

    def test_reading_attrs(self):
        """Test loading data."""
        from satpy.readers import load_reader
        filenames = [
            'W_XX-EUMETSAT-Darmstadt,SURFACE+SATELLITE,HY2B+SM_C_EUMP_20200326------_07077_o_250_l2b.h5', ]

        reader = load_reader(self.reader_configs)
        files = reader.select_files_from_pathnames(filenames)
        reader.create_filehandlers(files)
        # Make sure we have some files
        res = reader.load(['wvc_lon'])
        self.assertEqual(res['wvc_lon'].attrs['L2B_Number_WVC_cells'], 10)
        with self.assertRaises(KeyError):
            self.assertEqual(res['wvc_lon'].attrs['L2B_Expected_WVC_Cells'], 10)

    def test_reading_attrs_nsoas(self):
        """Test loading data."""
        from satpy.readers import load_reader
        filenames = [
            'H2B_OPER_SCA_L2B_OR_20210803T100304_20210803T104601_13905_pwp_250_07_owv.h5', ]

        reader = load_reader(self.reader_configs)
        files = reader.select_files_from_pathnames(filenames)
        reader.create_filehandlers(files)
        # Make sure we have some files
        res = reader.load(['wvc_lon'])
        with self.assertRaises(KeyError):
            self.assertEqual(res['wvc_lon'].attrs['L2B_Number_WVC_cells'], 10)
        self.assertEqual(res['wvc_lon'].attrs['L2B_Expected_WVC_Cells'], 10)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""SEVIRI L2 BUFR format reader.

References:
    EUMETSAT Product Navigator
    https://navigator.eumetsat.int/

"""
import logging
from datetime import datetime, timedelta

import dask.array as da
import numpy as np
import xarray as xr

from satpy.readers.eum_base import recarray2dict
from satpy.readers.seviri_base import mpef_product_header

try:
    import eccodes as ec
except ImportError:
    raise ImportError(
        "Missing eccodes-python and/or eccodes C-library installation. Use conda to install eccodes")

from satpy import CHUNK_SIZE
from satpy.readers.file_handlers import BaseFileHandler

logger = logging.getLogger('SeviriL2Bufr')

data_center_dict = {55: {'ssp': 'E0415', 'name': '08'}, 56:  {'ssp': 'E0000', 'name': '09'},
                    57: {'ssp': 'E0095', 'name': '10'}, 70: {'ssp': 'E0000', 'name': '11'}}

seg_size_dict = {'seviri_l2_bufr_asr': 16, 'seviri_l2_bufr_cla': 16,
                 'seviri_l2_bufr_csr': 16, 'seviri_l2_bufr_gii': 3,
                 'seviri_l2_bufr_thu': 16, 'seviri_l2_bufr_toz': 3}


class SeviriL2BufrFileHandler(BaseFileHandler):
    """File handler for SEVIRI L2 BUFR products."""

    def __init__(self, filename, filename_info, filetype_info, **kwargs):
        """Initialise the file handler for SEVIRI L2 BUFR data."""
        super(SeviriL2BufrFileHandler, self).__init__(filename,
                                                      filename_info,
                                                      filetype_info)

        if ('server' in filename_info):
            # EUMETSAT Offline Bufr product
            self.mpef_header = self._read_mpef_header()
        else:
            # Product was retrieved from the EUMETSAT Data Center
            timeStr = self.get_attribute('typicalDate')+self.get_attribute('typicalTime')
            buf_start_time = datetime.strptime(timeStr, "%Y%m%d%H%M%S")
            sc_id = self.get_attribute('satelliteIdentifier')
            self.mpef_header = {}
            self.mpef_header['NominalTime'] = buf_start_time
            self.mpef_header['SpacecraftName'] = data_center_dict[sc_id]['name']
            self.mpef_header['RectificationLongitude'] = data_center_dict[sc_id]['ssp']

        self.seg_size = seg_size_dict[filetype_info['file_type']]

    @property
    def start_time(self):
        """Return the repeat cycle start time."""
        return self.mpef_header['NominalTime']

    @property
    def end_time(self):
        """Return the repeat cycle end time."""
        return self.start_time + timedelta(minutes=15)

    @property
    def platform_name(self):
        """Return spacecraft name."""
        return 'MET{}'.format(self.mpef_header['SpacecraftName'])

    @property
    def ssp_lon(self):
        """Return subsatellite point longitude."""
        # e.g. E0415
        ssp_lon = self.mpef_header['RectificationLongitude']
        return float(ssp_lon[1:])/10.

    def _read_mpef_header(self):
        """Read MPEF header."""
        hdr = np.fromfile(self.filename, mpef_product_header, 1)
        return recarray2dict(hdr)

    def get_attribute(self, key):
        """Get BUFR attributes."""
        # This function is inefficient as it is looping through the entire
        # file to get 1 attribute. It causes a problem though if you break
        # from the file early - dont know why but investigating - fix later
        fh = open(self.filename, "rb")
        while True:
            # get handle for message
            bufr = ec.codes_bufr_new_from_file(fh)
            if bufr is None:
                break
            ec.codes_set(bufr, 'unpack', 1)
            attr = ec.codes_get(bufr, key)
            ec.codes_release(bufr)

        fh.close()
        return attr

    def get_array(self, key):
        """Get all data from file for the given BUFR key."""
        with open(self.filename, "rb") as fh:
            msgCount = 0
            while True:
                bufr = ec.codes_bufr_new_from_file(fh)
                if bufr is None:
                    break

                ec.codes_set(bufr, 'unpack', 1)

                # if is the first message initialise our final array
                if (msgCount == 0):
                    arr = da.from_array(ec.codes_get_array(
                        bufr, key, float), chunks=CHUNK_SIZE)
                else:
                    tmpArr = da.from_array(ec.codes_get_array(
                        bufr, key, float), chunks=CHUNK_SIZE)
                    arr = da.concatenate((arr, tmpArr))

                msgCount = msgCount+1
                ec.codes_release(bufr)

        if arr.size == 1:
            arr = arr[0]

        return arr

    def get_dataset(self, dataset_id, dataset_info):
        """Get dataset using the BUFR key in dataset_info."""
        arr = self.get_array(dataset_info['key'])
        arr[arr == dataset_info['fill_value']] = np.nan

        xarr = xr.DataArray(arr, dims=["y"])
        xarr.attrs['sensor'] = 'SEVIRI'
        xarr.attrs['platform_name'] = self.platform_name
        xarr.attrs['ssp_lon'] = self.ssp_lon
        xarr.attrs['seg_size'] = self.seg_size
        xarr.attrs.update(dataset_info)

        return xarr



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2020 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for SAR compositors."""

import unittest


class TestSARComposites(unittest.TestCase):
    """Test SAR-specific composites."""

    def test_sar_ice(self):
        """Test creating a the sar_ice composite."""
        import dask.array as da
        import numpy as np
        import xarray as xr

        from satpy.composites.sar import SARIce

        rows = 2
        cols = 2
        comp = SARIce('sar_ice', prerequisites=('hh', 'hv'),
                      standard_name='sar-ice')
        hh = xr.DataArray(da.zeros((rows, cols), chunks=25) + 2000,
                          dims=('y', 'x'),
                          attrs={'name': 'hh'})
        hv = xr.DataArray(da.zeros((rows, cols), chunks=25) + 1000,
                          dims=('y', 'x'),
                          attrs={'name': 'hv'})

        res = comp((hh, hv))
        self.assertIsInstance(res, xr.DataArray)
        self.assertIsInstance(res.data, da.Array)
        self.assertEqual(res.attrs['name'], 'sar_ice')
        self.assertEqual(res.attrs['standard_name'],
                         'sar-ice')
        data = res.compute()
        np.testing.assert_allclose(data.sel(bands='R'), 31.58280822)
        np.testing.assert_allclose(data.sel(bands='G'), 159869.56789876)
        np.testing.assert_allclose(data.sel(bands='B'), 44.68138191)

    def test_sar_ice_log(self):
        """Test creating a the sar_ice_log composite."""
        import dask.array as da
        import numpy as np
        import xarray as xr

        from satpy.composites.sar import SARIceLog

        rows = 2
        cols = 2
        comp = SARIceLog('sar_ice_log', prerequisites=('hh', 'hv'),
                         standard_name='sar-ice-log')
        hh = xr.DataArray(da.zeros((rows, cols), chunks=25) - 10,
                          dims=('y', 'x'),
                          attrs={'name': 'hh'})
        hv = xr.DataArray(da.zeros((rows, cols), chunks=25) - 20,
                          dims=('y', 'x'),
                          attrs={'name': 'hv'})

        res = comp((hh, hv))
        self.assertIsInstance(res, xr.DataArray)
        self.assertIsInstance(res.data, da.Array)
        self.assertEqual(res.attrs['name'], 'sar_ice_log')
        self.assertEqual(res.attrs['standard_name'],
                         'sar-ice-log')
        data = res.compute()
        np.testing.assert_allclose(data.sel(bands='R'), -20)
        np.testing.assert_allclose(data.sel(bands='G'), -4.6)
        np.testing.assert_allclose(data.sel(bands='B'), -10)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2020 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""The ahi_l1b_gridded_bin reader tests package."""

import os
import shutil
import tempfile
import unittest
from unittest import mock

import dask.array as da
import numpy as np
from pyresample.geometry import AreaDefinition

from satpy.readers.ahi_l1b_gridded_bin import AHI_LUT_NAMES, AHIGriddedFileHandler


class TestAHIGriddedArea(unittest.TestCase):
    """Test the AHI gridded reader definition."""

    def setUp(self):
        """Create fake data for testing."""
        self.FULLDISK_SIZES = {0.005: {'x_size': 24000,
                                       'y_size': 24000},
                               0.01: {'x_size': 12000,
                                      'y_size': 12000},
                               0.02: {'x_size': 6000,
                                      'y_size': 6000}}

        self.AHI_FULLDISK_EXTENT = [85., -60., 205., 60.]

    @staticmethod
    def make_fh(filetype, area='fld'):
        """Create a test file handler."""
        m = mock.mock_open()
        with mock.patch('satpy.readers.ahi_l1b_gridded_bin.open', m, create=True):
            fh = AHIGriddedFileHandler('somefile',
                                       {'area': area},
                                       filetype_info={'file_type': filetype})
            return fh

    def test_low_res(self):
        """Check size of the low resolution (2km) grid."""
        tmp_fh = self.make_fh('tir.01')
        self.assertEqual(self.FULLDISK_SIZES[0.02]['x_size'], tmp_fh.ncols)
        self.assertEqual(self.FULLDISK_SIZES[0.02]['y_size'], tmp_fh.nlines)

    def test_med_res(self):
        """Check size of the low resolution (1km) grid."""
        tmp_fh = self.make_fh('vis.02')
        self.assertEqual(self.FULLDISK_SIZES[0.01]['x_size'], tmp_fh.ncols)
        self.assertEqual(self.FULLDISK_SIZES[0.01]['y_size'], tmp_fh.nlines)

    def test_hi_res(self):
        """Check size of the low resolution (0.5km) grid."""
        tmp_fh = self.make_fh('ext.01')
        self.assertEqual(self.FULLDISK_SIZES[0.005]['x_size'], tmp_fh.ncols)
        self.assertEqual(self.FULLDISK_SIZES[0.005]['y_size'], tmp_fh.nlines)

    def test_area_def(self):
        """Check that a valid full disk area is produced."""
        good_area = AreaDefinition('gridded_himawari',
                                   'A gridded Himawari area',
                                   'longlat',
                                   'EPSG:4326',
                                   self.FULLDISK_SIZES[0.01]['x_size'],
                                   self.FULLDISK_SIZES[0.01]['y_size'],
                                   self.AHI_FULLDISK_EXTENT)

        tmp_fh = self.make_fh('vis.01')
        tmp_fh.get_area_def(None)
        self.assertEqual(tmp_fh.area, good_area)

    def test_bad_area(self):
        """Ensure an error is raised for an usupported area."""
        tmp_fh = self.make_fh('ext.01')
        tmp_fh.areaname = 'scanning'
        with self.assertRaises(NotImplementedError):
            tmp_fh.get_area_def(None)
        with self.assertRaises(NotImplementedError):
            self.make_fh('ext.01', area='scanning')


class TestAHIGriddedFileCalibration(unittest.TestCase):
    """Test case for the file calibration types."""

    def setUp(self):
        """Create a test file handler."""
        m = mock.mock_open()
        with mock.patch('satpy.readers.ahi_l1b_gridded_bin.open', m, create=True):
            in_fname = 'test_file'
            fh = AHIGriddedFileHandler(in_fname,
                                       {'area': 'fld'},
                                       filetype_info={'file_type': 'tir.01'})
            self.fh = fh

    @mock.patch('satpy.readers.ahi_l1b_gridded_bin.AHIGriddedFileHandler._get_luts')
    @mock.patch('satpy.readers.ahi_l1b_gridded_bin.os.path.exists')
    @mock.patch('satpy.readers.ahi_l1b_gridded_bin.np.loadtxt')
    def test_calibrate(self, np_loadtxt, os_exist, get_luts):
        """Test the calibration modes of AHI using the LUTs."""
        load_return = np.squeeze(np.dstack([np.arange(0, 2048, 1),
                                            np.arange(0, 120, 0.05859375)]))

        np_loadtxt.return_value = load_return
        get_luts.return_value = True

        in_data = np.array([[100., 300., 500.],
                           [800., 1500., 2040.]])

        refl_out = np.array([[5.859375, 17.578125, 29.296875],
                             [46.875, 87.890625, 119.53125]])

        os_exist.return_value = False
        # Check that the LUT download is called if we don't have the LUTS
        self.fh.calibrate(in_data, 'reflectance')
        get_luts.assert_called()

        os_exist.return_value = True
        # Ensure results equal if no calibration applied
        out_data = self.fh.calibrate(in_data, 'counts')
        np.testing.assert_equal(in_data, out_data)

        # Now ensure results equal if LUT calibration applied
        out_data = self.fh.calibrate(in_data, 'reflectance')
        np.testing.assert_allclose(refl_out, out_data)

        # Check that exception is raised if bad calibration is passed
        with self.assertRaises(NotImplementedError):
            self.fh.calibrate(in_data, 'lasers')

        # Check that exception is raised if no file is present
        np_loadtxt.side_effect = FileNotFoundError
        with self.assertRaises(FileNotFoundError):
            self.fh.calibrate(in_data, 'reflectance')


class TestAHIGriddedFileHandler(unittest.TestCase):
    """Test case for the file reading."""

    def new_unzip(fname):
        """Fake unzipping."""
        if fname[-3:] == 'bz2':
            return fname[:-4]

    @mock.patch('satpy.readers.ahi_l1b_gridded_bin.unzip_file',
                mock.MagicMock(side_effect=new_unzip))
    def setUp(self):
        """Create a test file handler."""
        m = mock.mock_open()
        with mock.patch('satpy.readers.ahi_l1b_gridded_bin.open', m, create=True):
            in_fname = 'test_file.bz2'
            fh = AHIGriddedFileHandler(in_fname,
                                       {'area': 'fld'},
                                       filetype_info={'file_type': 'tir.01'})

            # Check that the filename is altered for bz2 format files
            self.assertNotEqual(in_fname, fh.filename)
            self.fh = fh

        key = {'calibration': 'counts',
               'name': 'vis.01'}
        info = {'units': 'unitless',
                'standard_name': 'vis.01',
                'wavelength': 10.8,
                'resolution': 0.05}
        self.key = key
        self.info = info

    @mock.patch('satpy.readers.ahi_l1b_gridded_bin.np.memmap')
    def test_dataread(self, memmap):
        """Check that a dask array is returned from the read function."""
        test_arr = np.zeros((10, 10))
        memmap.return_value = test_arr
        m = mock.mock_open()
        with mock.patch('satpy.readers.ahi_l1b_gridded_bin.open', m, create=True):
            res = self.fh._read_data(mock.MagicMock())
            np.testing.assert_allclose(res, da.from_array(test_arr))

    @mock.patch('satpy.readers.ahi_l1b_gridded_bin.AHIGriddedFileHandler._read_data')
    def test_get_dataset(self, mocked_read):
        """Check that a good dataset is returned on request."""
        m = mock.mock_open()

        out_data = np.array([[100., 300., 500.],
                             [800., 1500., 2040.]])
        mocked_read.return_value = out_data
        with mock.patch('satpy.readers.ahi_l1b_gridded_bin.open', m, create=True):
            res = self.fh.get_dataset(self.key, self.info)
            mocked_read.assert_called()
            # Check output data is correct
            np.testing.assert_allclose(res.values, out_data)
            # Also check a couple of attributes
            self.assertEqual(res.attrs['name'], self.key['name'])
            self.assertEqual(res.attrs['wavelength'], self.info['wavelength'])

    @mock.patch('os.path.exists', return_value=True)
    @mock.patch('os.remove')
    def test_destructor(self, exist_patch, remove_patch):
        """Check that file handler deletes files if needed."""
        del self.fh
        remove_patch.assert_called()


class TestAHIGriddedLUTs(unittest.TestCase):
    """Test case for the downloading and preparing LUTs."""

    def mocked_ftp_dl(fname):
        """Fake download of LUT tar file by creating a local tar."""
        import os
        import tarfile
        import tempfile

        with tarfile.open(fname, "w:gz") as tar_handle:
            for namer in AHI_LUT_NAMES:
                tmpf = os.path.join(tempfile.tempdir, namer)
                with open(tmpf, 'w') as tmp_fid:
                    tmp_fid.write("TEST\n")
                tar_handle.add(tmpf, arcname='count2tbb_v102/'+namer)
                os.remove(tmpf)

    def setUp(self):
        """Create a test file handler."""
        m = mock.mock_open()
        with mock.patch('satpy.readers.ahi_l1b_gridded_bin.open', m, create=True):
            in_fname = 'test_file'
            fh = AHIGriddedFileHandler(in_fname,
                                       {'area': 'fld'},
                                       filetype_info={'file_type': 'tir.01'})
            self.fh = fh

        key = {'calibration': 'counts',
               'name': 'vis.01'}
        info = {'units': 'unitless',
                'standard_name': 'vis.01',
                'wavelength': 10.8,
                'resolution': 0.05}
        self.key = key
        self.info = info

    def tearDown(self):
        """Remove files and directories created by the tests."""
        for lut_name in AHI_LUT_NAMES:
            tmp_filename = os.path.join(self.fh.lut_dir, lut_name)
            if os.path.isfile(tmp_filename):
                os.remove(tmp_filename)
        if os.path.isdir(self.fh.lut_dir):
            shutil.rmtree(self.fh.lut_dir)

    @mock.patch('satpy.readers.ahi_l1b_gridded_bin.AHIGriddedFileHandler._download_luts',
                mock.MagicMock(side_effect=mocked_ftp_dl))
    def test_get_luts(self):
        """Check that the function to download LUTs operates successfully."""
        tempdir = tempfile.gettempdir()
        print(self.fh.lut_dir)
        self.fh._get_luts()
        self.assertFalse(os.path.exists(os.path.join(tempdir, 'count2tbb_v102/')))
        for lut_name in AHI_LUT_NAMES:
            self.assertTrue(os.path.isfile(os.path.join(self.fh.lut_dir, lut_name)))

    @mock.patch('urllib.request.urlopen')
    @mock.patch('shutil.copyfileobj')
    def test_download_luts(self, mock_dl, mock_shutil):
        """Test that the FTP library is called for downloading LUTS."""
        m = mock.mock_open()
        with mock.patch('satpy.readers.ahi_l1b_gridded_bin.open', m, create=True):
            self.fh._download_luts('/test_file')
            mock_dl.assert_called()
            mock_shutil.assert_called()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2015-2021 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Composite classes for the AHI instrument."""

import logging

from satpy.composites import GenericCompositor
from satpy.dataset import combine_metadata

LOG = logging.getLogger(__name__)


class GreenCorrector(GenericCompositor):
    """Corrector of the AHI green band to compensate for the deficit of chlorophyll signal."""

    def __init__(self, *args, fractions=(0.85, 0.15), **kwargs):
        """Set default keyword argument values."""
        # XXX: Should this be 0.93 and 0.07
        self.fractions = fractions
        super(GreenCorrector, self).__init__(*args, **kwargs)

    def __call__(self, projectables, optional_datasets=None, **attrs):
        """Boost vegetation effect thanks to NIR (0.8µm) band."""
        LOG.info('Boosting vegetation on green band')

        projectables = self.match_data_arrays(projectables)
        new_green = sum(fraction * value for fraction, value in zip(self.fractions, projectables))
        new_green.attrs = combine_metadata(*projectables)
        return super(GreenCorrector, self).__call__((new_green,), **attrs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2014-2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Helper functions for satpy readers."""

import bz2
import logging
import os
import shutil
import tempfile
import warnings
from contextlib import closing
from io import BytesIO
from shutil import which
from subprocess import PIPE, Popen  # nosec

import numpy as np
import pyproj
import xarray as xr
from pyresample.geometry import AreaDefinition

from satpy import CHUNK_SIZE

LOGGER = logging.getLogger(__name__)


def np2str(value):
    """Convert an `numpy.string_` to str.

    Args:
        value (ndarray): scalar or 1-element numpy array to convert

    Raises:
        ValueError: if value is array larger than 1-element or it is not of
                    type `numpy.string_` or it is not a numpy array

    """
    if hasattr(value, 'dtype') and \
            issubclass(value.dtype.type, (np.str_, np.string_, np.object_)) \
            and value.size == 1:
        value = value.item()
        if not isinstance(value, str):
            # python 3 - was scalar numpy array of bytes
            # otherwise python 2 - scalar numpy array of 'str'
            value = value.decode()
        return value
    else:
        raise ValueError("Array is not a string type or is larger than 1")


def _get_geostationary_height(geos_area):
    params = geos_area.crs.coordinate_operation.params
    h_param = [p for p in params if 'satellite height' in p.name.lower()][0]
    return h_param.value


def _get_geostationary_reference_longitude(geos_area):
    params = geos_area.crs.coordinate_operation.params
    lon_0_params = [p for p in params if 'longitude of natural origin' in p.name.lower()]
    if not lon_0_params:
        return 0
    elif len(lon_0_params) != 1:
        raise ValueError("Not sure how to get reference longitude "
                         "information from AreaDefinition.")
    return lon_0_params[0].value


def _get_geostationary_semi_axes(geos_area):
    from pyresample.utils import proj4_radius_parameters
    return proj4_radius_parameters(geos_area.crs)


def get_geostationary_angle_extent(geos_area):
    """Get the max earth (vs space) viewing angles in x and y."""
    # TODO: take into account sweep_axis_angle parameter
    a, b = _get_geostationary_semi_axes(geos_area)
    h = _get_geostationary_height(geos_area)
    req = float(a) / 1000
    rp = float(b) / 1000
    h = float(h) / 1000 + req

    # compute some constants
    aeq = 1 - req**2 / (h ** 2)
    ap_ = 1 - rp**2 / (h ** 2)

    # generate points around the north hemisphere in satellite projection
    # make it a bit smaller so that we stay inside the valid area
    xmax = np.arccos(np.sqrt(aeq))
    ymax = np.arccos(np.sqrt(ap_))
    return xmax, ymax


def get_geostationary_mask(area, chunks=None):
    """Compute a mask of the earth's shape as seen by a geostationary satellite.

    Args:
        area (pyresample.geometry.AreaDefinition) : Corresponding area
                                                    definition
        chunks (int or tuple): Chunk size for the 2D array that is generated.

    Returns:
        Boolean mask, True inside the earth's shape, False outside.

    """
    # Compute projection coordinates at the earth's limb
    h = _get_geostationary_height(area)
    xmax, ymax = get_geostationary_angle_extent(area)
    xmax *= h
    ymax *= h

    # Compute projection coordinates at the centre of each pixel
    x, y = area.get_proj_coords(chunks=chunks or CHUNK_SIZE)

    # Compute mask of the earth's elliptical shape
    return ((x / xmax) ** 2 + (y / ymax) ** 2) <= 1


def _lonlat_from_geos_angle(x, y, geos_area):
    """Get lons and lats from x, y in projection coordinates."""
    a, b = _get_geostationary_semi_axes(geos_area)
    h = _get_geostationary_height(geos_area)
    lon_0 = _get_geostationary_reference_longitude(geos_area)
    h__ = float(h + a) / 1000
    b__ = (a / float(b)) ** 2

    sd = np.sqrt((h__ * np.cos(x) * np.cos(y)) ** 2 -
                 (np.cos(y)**2 + b__ * np.sin(y)**2) *
                 (h__**2 - (float(a) / 1000)**2))
    # sd = 0

    sn = (h__ * np.cos(x) * np.cos(y) - sd) / (np.cos(y)**2 + b__ * np.sin(y)**2)
    s1 = h__ - sn * np.cos(x) * np.cos(y)
    s2 = sn * np.sin(x) * np.cos(y)
    s3 = -sn * np.sin(y)
    sxy = np.sqrt(s1**2 + s2**2)

    lons = np.rad2deg(np.arctan2(s2, s1)) + lon_0
    lats = np.rad2deg(-np.arctan2(b__ * s3, sxy))

    return lons, lats


def get_geostationary_bounding_box(geos_area, nb_points=50):
    """Get the bbox in lon/lats of the valid pixels inside *geos_area*.

    Args:
      geos_area: The geostationary area to analyse.
      nb_points: Number of points on the polygon

    """
    xmax, ymax = get_geostationary_angle_extent(geos_area)
    h = _get_geostationary_height(geos_area)

    # generate points around the north hemisphere in satellite projection
    # make it a bit smaller so that we stay inside the valid area
    x = np.cos(np.linspace(-np.pi, 0, nb_points // 2)) * (xmax - 0.001)
    y = -np.sin(np.linspace(-np.pi, 0, nb_points // 2)) * (ymax - 0.001)

    # clip the projection coordinates to fit the area extent of geos_area
    ll_x, ll_y, ur_x, ur_y = (np.array(geos_area.area_extent) /
                              float(h))

    x = np.clip(np.concatenate([x, x[::-1]]), min(ll_x, ur_x), max(ll_x, ur_x))
    y = np.clip(np.concatenate([y, -y]), min(ll_y, ur_y), max(ll_y, ur_y))

    return _lonlat_from_geos_angle(x, y, geos_area)


def get_sub_area(area, xslice, yslice):
    """Apply slices to the area_extent and size of the area."""
    new_area_extent = ((area.pixel_upper_left[0] +
                        (xslice.start - 0.5) * area.pixel_size_x),
                       (area.pixel_upper_left[1] -
                        (yslice.stop - 0.5) * area.pixel_size_y),
                       (area.pixel_upper_left[0] +
                        (xslice.stop - 0.5) * area.pixel_size_x),
                       (area.pixel_upper_left[1] -
                        (yslice.start - 0.5) * area.pixel_size_y))

    return AreaDefinition(area.area_id, area.name,
                          area.proj_id, area.crs,
                          xslice.stop - xslice.start,
                          yslice.stop - yslice.start,
                          new_area_extent)


def unzip_file(filename):
    """Unzip the file if file is bzipped = ending with 'bz2'."""
    if filename.endswith('bz2'):
        fdn, tmpfilepath = tempfile.mkstemp()
        LOGGER.info("Using temp file for BZ2 decompression: %s", tmpfilepath)
        # try pbzip2
        pbzip = which('pbzip2')
        # Run external pbzip2
        if pbzip is not None:
            n_thr = os.environ.get('OMP_NUM_THREADS')
            if n_thr:
                runner = [pbzip,
                          '-dc',
                          '-p'+str(n_thr),
                          filename]
            else:
                runner = [pbzip,
                          '-dc',
                          filename]
            p = Popen(runner, stdout=PIPE, stderr=PIPE)  # nosec
            stdout = BytesIO(p.communicate()[0])
            status = p.returncode
            if status != 0:
                raise IOError("pbzip2 error '%s', failed, status=%d"
                              % (filename, status))
            with closing(os.fdopen(fdn, 'wb')) as ofpt:
                try:
                    stdout.seek(0)
                    shutil.copyfileobj(stdout, ofpt)
                except IOError:
                    import traceback
                    traceback.print_exc()
                    LOGGER.info("Failed to read bzipped file %s",
                                str(filename))
                    os.remove(tmpfilepath)
                    raise
            return tmpfilepath

        # Otherwise, fall back to the original method
        bz2file = bz2.BZ2File(filename)
        with closing(os.fdopen(fdn, 'wb')) as ofpt:
            try:
                ofpt.write(bz2file.read())
            except IOError:
                import traceback
                traceback.print_exc()
                LOGGER.info("Failed to read bzipped file %s", str(filename))
                os.remove(tmpfilepath)
                return None
        return tmpfilepath

    return None


class unzip_context():
    """Context manager for uncompressing a .bz2 file on the fly.

    Uses `unzip_file`. Removes the uncompressed file on exit of the context manager.

    Returns: the filename of the uncompressed file or of the original file if it was not
    compressed.

    """

    def __init__(self, filename):
        """Keep original filename."""
        self.input_filename = filename

    def __enter__(self):
        """Uncompress file if necessary and return the relevant filename for the file handler."""
        unzipped = unzip_file(self.input_filename)
        if unzipped is not None:
            self.unzipped_filename = unzipped
            return unzipped
        else:
            self.unzipped_filename = None
            return self.input_filename

    def __exit__(self, exc_type, exc_value, traceback):
        """Remove temporary file."""
        if self.unzipped_filename is not None:
            os.remove(self.unzipped_filename)


def bbox(img):
    """Find the bounding box around nonzero elements in the given array.

    Copied from https://stackoverflow.com/a/31402351/5703449 .

    Returns:
        rowmin, rowmax, colmin, colmax

    """
    rows = np.any(img, axis=1)
    cols = np.any(img, axis=0)
    rmin, rmax = np.where(rows)[0][[0, -1]]
    cmin, cmax = np.where(cols)[0][[0, -1]]

    return rmin, rmax, cmin, cmax


def get_earth_radius(lon, lat, a, b):
    """Compute radius of the earth ellipsoid at the given longitude and latitude.

    Args:
        lon: Geodetic longitude (degrees)
        lat: Geodetic latitude (degrees)
        a: Semi-major axis of the ellipsoid (meters)
        b: Semi-minor axis of the ellipsoid (meters)

    Returns:
        Earth Radius (meters)

    """
    geocent = pyproj.Proj(proj='geocent', a=a, b=b, units='m')
    latlong = pyproj.Proj(proj='latlong', a=a, b=b, units='m')
    x, y, z = pyproj.transform(latlong, geocent, lon, lat, 0.)
    return np.sqrt(x**2 + y**2 + z**2)


def reduce_mda(mda, max_size=100):
    """Recursively remove arrays with more than `max_size` elements from the given metadata dictionary."""
    reduced = {}
    for key, val in mda.items():
        if isinstance(val, dict):
            reduced[key] = reduce_mda(val, max_size)
        elif not (isinstance(val, np.ndarray) and val.size > max_size):
            reduced[key] = val
    return reduced


def get_user_calibration_factors(band_name, correction_dict):
    """Retrieve radiance correction factors from user-supplied dict."""
    if band_name in correction_dict:
        try:
            slope = correction_dict[band_name]['slope']
            offset = correction_dict[band_name]['offset']
        except KeyError:
            raise KeyError("Incorrect correction factor dictionary. You must "
                           "supply 'slope' and 'offset' keys.")
    else:
        # If coefficients not present, warn user and use slope=1, offset=0
        warnings.warn("WARNING: You have selected radiance correction but "
                      " have not supplied coefficients for channel " +
                      band_name)
        return 1., 0.

    return slope, offset


def apply_rad_correction(data, slope, offset):
    """Apply GSICS-like correction factors to radiance data."""
    data = (data - offset) / slope
    return data


def get_array_date(scn_data, utc_date=None):
    """Get start time from a channel data array."""
    if utc_date is None:
        try:
            utc_date = scn_data.attrs['start_time']
        except KeyError:
            try:
                utc_date = scn_data.attrs['scheduled_time']
            except KeyError:
                raise KeyError('Scene has no start_time '
                               'or scheduled_time attribute.')
    return utc_date


def apply_earthsun_distance_correction(reflectance, utc_date=None):
    """Correct reflectance data to account for changing Earth-Sun distance."""
    from pyorbital.astronomy import sun_earth_distance_correction
    utc_date = get_array_date(reflectance, utc_date)
    sun_earth_dist = sun_earth_distance_correction(utc_date)

    reflectance.attrs['sun_earth_distance_correction_applied'] = True
    reflectance.attrs['sun_earth_distance_correction_factor'] = sun_earth_dist
    with xr.set_options(keep_attrs=True):
        reflectance = reflectance * sun_earth_dist * sun_earth_dist
    return reflectance


def remove_earthsun_distance_correction(reflectance, utc_date=None):
    """Remove the sun-earth distance correction."""
    from pyorbital.astronomy import sun_earth_distance_correction
    utc_date = get_array_date(reflectance, utc_date)
    sun_earth_dist = sun_earth_distance_correction(utc_date)

    reflectance.attrs['sun_earth_distance_correction_applied'] = False
    reflectance.attrs['sun_earth_distance_correction_factor'] = sun_earth_dist
    with xr.set_options(keep_attrs=True):
        reflectance = reflectance / (sun_earth_dist * sun_earth_dist)
    return reflectance



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-2018 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Module for testing the satpy.readers.viirs_l1b module."""

import os
import unittest
from datetime import datetime, timedelta
from unittest import mock

import numpy as np

from satpy.tests.reader_tests.test_netcdf_utils import FakeNetCDF4FileHandler
from satpy.tests.utils import convert_file_content_to_data_array

DEFAULT_FILE_DTYPE = np.uint16
DEFAULT_FILE_SHAPE = (10, 300)
DEFAULT_FILE_DATA = np.arange(DEFAULT_FILE_SHAPE[0] * DEFAULT_FILE_SHAPE[1],
                              dtype=DEFAULT_FILE_DTYPE).reshape(DEFAULT_FILE_SHAPE)
DEFAULT_FILE_FACTORS = np.array([2.0, 1.0], dtype=np.float32)
DEFAULT_LAT_DATA = np.linspace(45, 65, DEFAULT_FILE_SHAPE[1]).astype(DEFAULT_FILE_DTYPE)
DEFAULT_LAT_DATA = np.repeat([DEFAULT_LAT_DATA], DEFAULT_FILE_SHAPE[0], axis=0)
DEFAULT_LON_DATA = np.linspace(5, 45, DEFAULT_FILE_SHAPE[1]).astype(DEFAULT_FILE_DTYPE)
DEFAULT_LON_DATA = np.repeat([DEFAULT_LON_DATA], DEFAULT_FILE_SHAPE[0], axis=0)


class FakeNetCDF4FileHandler2(FakeNetCDF4FileHandler):
    """Swap-in NetCDF4 File Handler."""

    def get_test_content(self, filename, filename_info, filetype_info):
        """Mimic reader input file content."""
        dt = filename_info.get('start_time', datetime(2016, 1, 1, 12, 0, 0))
        file_type = filename[:5].lower()
        # num_lines = {
        #     'vl1bi': 3248 * 2,
        #     'vl1bm': 3248,
        #     'vl1bd': 3248,
        # }[file_type]
        # num_pixels = {
        #     'vl1bi': 6400,
        #     'vl1bm': 3200,
        #     'vl1bd': 4064,
        # }[file_type]
        # num_scans = 203
        # num_luts = 65536
        num_lines = DEFAULT_FILE_SHAPE[0]
        num_pixels = DEFAULT_FILE_SHAPE[1]
        num_scans = 5
        num_luts = DEFAULT_FILE_SHAPE[0] * DEFAULT_FILE_SHAPE[1]
        file_content = {
            '/dimension/number_of_scans': num_scans,
            '/dimension/number_of_lines': num_lines,
            '/dimension/number_of_pixels': num_pixels,
            '/dimension/number_of_LUT_values': num_luts,
            '/attr/time_coverage_start': dt.strftime('%Y-%m-%dT%H:%M:%S.000Z'),
            '/attr/time_coverage_end': (dt + timedelta(minutes=6)).strftime('%Y-%m-%dT%H:%M:%S.000Z'),
            '/attr/orbit_number': 26384,
            '/attr/instrument': 'viirs',
            '/attr/platform': 'Suomi-NPP',
        }
        self._fill_contents_with_default_data(file_content, file_type)

        self._set_dataset_specific_metadata(file_content)

        convert_file_content_to_data_array(file_content)
        return file_content

    @staticmethod
    def _fill_contents_with_default_data(file_content, file_type):
        """Fill file contents with default data."""
        if file_type.startswith('vgeo'):
            file_content['/attr/OrbitNumber'] = file_content.pop('/attr/orbit_number')
            file_content['geolocation_data/latitude'] = DEFAULT_LAT_DATA
            file_content['geolocation_data/longitude'] = DEFAULT_LON_DATA
            file_content['geolocation_data/solar_zenith'] = DEFAULT_LON_DATA
            file_content['geolocation_data/solar_azimuth'] = DEFAULT_LON_DATA
            file_content['geolocation_data/sensor_zenith'] = DEFAULT_LON_DATA
            file_content['geolocation_data/sensor_azimuth'] = DEFAULT_LON_DATA
            if file_type.endswith('d'):
                file_content['geolocation_data/lunar_zenith'] = DEFAULT_LON_DATA
                file_content['geolocation_data/lunar_azimuth'] = DEFAULT_LON_DATA
        elif file_type == 'vl1bm':
            file_content['observation_data/M01'] = DEFAULT_FILE_DATA
            file_content['observation_data/M02'] = DEFAULT_FILE_DATA
            file_content['observation_data/M03'] = DEFAULT_FILE_DATA
            file_content['observation_data/M04'] = DEFAULT_FILE_DATA
            file_content['observation_data/M05'] = DEFAULT_FILE_DATA
            file_content['observation_data/M06'] = DEFAULT_FILE_DATA
            file_content['observation_data/M07'] = DEFAULT_FILE_DATA
            file_content['observation_data/M08'] = DEFAULT_FILE_DATA
            file_content['observation_data/M09'] = DEFAULT_FILE_DATA
            file_content['observation_data/M10'] = DEFAULT_FILE_DATA
            file_content['observation_data/M11'] = DEFAULT_FILE_DATA
            file_content['observation_data/M12'] = DEFAULT_FILE_DATA
            file_content['observation_data/M13'] = DEFAULT_FILE_DATA
            file_content['observation_data/M14'] = DEFAULT_FILE_DATA
            file_content['observation_data/M15'] = DEFAULT_FILE_DATA
            file_content['observation_data/M16'] = DEFAULT_FILE_DATA
        elif file_type == 'vl1bi':
            file_content['observation_data/I01'] = DEFAULT_FILE_DATA
            file_content['observation_data/I02'] = DEFAULT_FILE_DATA
            file_content['observation_data/I03'] = DEFAULT_FILE_DATA
            file_content['observation_data/I04'] = DEFAULT_FILE_DATA
            file_content['observation_data/I05'] = DEFAULT_FILE_DATA
        elif file_type == 'vl1bd':
            file_content['observation_data/DNB_observations'] = DEFAULT_FILE_DATA
            file_content['observation_data/DNB_observations/attr/units'] = 'Watts/cm^2/steradian'

    @staticmethod
    def _set_dataset_specific_metadata(file_content):
        """Set dataset-specific metadata."""
        for k in list(file_content.keys()):
            if not k.startswith('observation_data') and not k.startswith('geolocation_data'):
                continue
            file_content[k + '/shape'] = DEFAULT_FILE_SHAPE
            if k[-3:] in ['M12', 'M13', 'M14', 'M15', 'M16', 'I04', 'I05']:
                file_content[k + '_brightness_temperature_lut'] = DEFAULT_FILE_DATA.ravel()
                file_content[k + '_brightness_temperature_lut/attr/units'] = 'Kelvin'
                file_content[k + '_brightness_temperature_lut/attr/valid_min'] = 0
                file_content[k + '_brightness_temperature_lut/attr/valid_max'] = 65534
                file_content[k + '_brightness_temperature_lut/attr/_FillValue'] = 65535
                file_content[k + '/attr/units'] = 'Watts/meter^2/steradian/micrometer'
            elif k[-3:] in ['M01', 'M02', 'M03', 'M04', 'M05', 'M06', 'M07', 'M08',
                            'M09', 'M10', 'M11', 'I01', 'I02', 'I03']:
                file_content[k + '/attr/radiance_units'] = 'Watts/meter^2/steradian/micrometer'
                file_content[k + '/attr/radiance_scale_factor'] = 1.1
                file_content[k + '/attr/radiance_add_offset'] = 0.1
            elif k.endswith('longitude'):
                file_content[k + '/attr/units'] = 'degrees_east'
            elif k.endswith('latitude'):
                file_content[k + '/attr/units'] = 'degrees_north'
            elif k.endswith('zenith') or k.endswith('azimuth'):
                file_content[k + '/attr/units'] = 'degrees'
            file_content[k + '/attr/valid_min'] = 0
            file_content[k + '/attr/valid_max'] = 65534
            file_content[k + '/attr/_FillValue'] = 65535
            file_content[k + '/attr/scale_factor'] = 1.1
            file_content[k + '/attr/add_offset'] = 0.1


class TestVIIRSL1BReader(unittest.TestCase):
    """Test VIIRS L1B Reader."""

    yaml_file = "viirs_l1b.yaml"

    def setUp(self):
        """Wrap NetCDF4 file handler with our own fake handler."""
        from satpy._config import config_search_paths
        from satpy.readers.viirs_l1b import VIIRSL1BFileHandler
        self.reader_configs = config_search_paths(os.path.join('readers', self.yaml_file))
        # http://stackoverflow.com/questions/12219967/how-to-mock-a-base-class-with-python-mock-library
        self.p = mock.patch.object(VIIRSL1BFileHandler, '__bases__', (FakeNetCDF4FileHandler2,))
        self.fake_handler = self.p.start()
        self.p.is_local = True

    def tearDown(self):
        """Stop wrapping the NetCDF4 file handler."""
        self.p.stop()

    def test_init(self):
        """Test basic init with no extra parameters."""
        from satpy.readers import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            'VL1BM_snpp_d20161130_t012400_c20161130054822.nc',
        ])
        self.assertEqual(len(loadables), 1)
        r.create_filehandlers(loadables)
        # make sure we have some files
        self.assertTrue(r.file_handlers)

    def test_load_every_m_band_bt(self):
        """Test loading all M band brightness temperatures."""
        from satpy.readers import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            'VL1BM_snpp_d20161130_t012400_c20161130054822.nc',
            'VGEOM_snpp_d20161130_t012400_c20161130054822.nc',
        ])
        r.create_filehandlers(loadables)
        datasets = r.load(['M12',
                           'M13',
                           'M14',
                           'M15',
                           'M16'])
        self.assertEqual(len(datasets), 5)
        for v in datasets.values():
            self.assertEqual(v.attrs['calibration'], 'brightness_temperature')
            self.assertEqual(v.attrs['units'], 'K')
            self.assertEqual(v.attrs['rows_per_scan'], 2)
            self.assertEqual(v.attrs['area'].lons.attrs['rows_per_scan'], 2)
            self.assertEqual(v.attrs['area'].lats.attrs['rows_per_scan'], 2)

    def test_load_every_m_band_refl(self):
        """Test loading all M band reflectances."""
        from satpy.readers import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            'VL1BM_snpp_d20161130_t012400_c20161130054822.nc',
            'VGEOM_snpp_d20161130_t012400_c20161130054822.nc',
        ])
        r.create_filehandlers(loadables)
        datasets = r.load(['M01',
                           'M02',
                           'M03',
                           'M04',
                           'M05',
                           'M06',
                           'M07',
                           'M08',
                           'M09',
                           'M10',
                           'M11'])
        self.assertEqual(len(datasets), 11)
        for v in datasets.values():
            self.assertEqual(v.attrs['calibration'], 'reflectance')
            self.assertEqual(v.attrs['units'], '%')
            self.assertEqual(v.attrs['rows_per_scan'], 2)
            self.assertEqual(v.attrs['area'].lons.attrs['rows_per_scan'], 2)
            self.assertEqual(v.attrs['area'].lats.attrs['rows_per_scan'], 2)

    def test_load_every_m_band_rad(self):
        """Test loading all M bands as radiances."""
        from satpy.readers import load_reader
        from satpy.tests.utils import make_dataid
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            'VL1BM_snpp_d20161130_t012400_c20161130054822.nc',
            'VGEOM_snpp_d20161130_t012400_c20161130054822.nc',
        ])
        r.create_filehandlers(loadables)
        datasets = r.load([make_dataid(name='M01', calibration='radiance'),
                           make_dataid(name='M02', calibration='radiance'),
                           make_dataid(name='M03', calibration='radiance'),
                           make_dataid(name='M04', calibration='radiance'),
                           make_dataid(name='M05', calibration='radiance'),
                           make_dataid(name='M06', calibration='radiance'),
                           make_dataid(name='M07', calibration='radiance'),
                           make_dataid(name='M08', calibration='radiance'),
                           make_dataid(name='M09', calibration='radiance'),
                           make_dataid(name='M10', calibration='radiance'),
                           make_dataid(name='M11', calibration='radiance'),
                           make_dataid(name='M12', calibration='radiance'),
                           make_dataid(name='M13', calibration='radiance'),
                           make_dataid(name='M14', calibration='radiance'),
                           make_dataid(name='M15', calibration='radiance'),
                           make_dataid(name='M16', calibration='radiance')])
        self.assertEqual(len(datasets), 16)
        for v in datasets.values():
            self.assertEqual(v.attrs['calibration'], 'radiance')
            self.assertEqual(v.attrs['units'], 'W m-2 um-1 sr-1')
            self.assertEqual(v.attrs['rows_per_scan'], 2)
            self.assertEqual(v.attrs['area'].lons.attrs['rows_per_scan'], 2)
            self.assertEqual(v.attrs['area'].lats.attrs['rows_per_scan'], 2)

    def test_load_dnb_radiance(self):
        """Test loading the main DNB dataset."""
        from satpy.readers import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            'VL1BD_snpp_d20161130_t012400_c20161130054822.nc',
            'VGEOD_snpp_d20161130_t012400_c20161130054822.nc',
        ])
        r.create_filehandlers(loadables)
        datasets = r.load(['DNB'])
        self.assertEqual(len(datasets), 1)
        for v in datasets.values():
            self.assertEqual(v.attrs['calibration'], 'radiance')
            self.assertEqual(v.attrs['units'], 'W m-2 sr-1')
            self.assertEqual(v.attrs['rows_per_scan'], 2)
            self.assertEqual(v.attrs['area'].lons.attrs['rows_per_scan'], 2)
            self.assertEqual(v.attrs['area'].lats.attrs['rows_per_scan'], 2)

    def test_load_dnb_angles(self):
        """Test loading all DNB angle datasets."""
        from satpy.readers import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            'VL1BD_snpp_d20161130_t012400_c20161130054822.nc',
            'VGEOD_snpp_d20161130_t012400_c20161130054822.nc',
        ])
        r.create_filehandlers(loadables)
        datasets = r.load(['dnb_solar_zenith_angle',
                           'dnb_solar_azimuth_angle',
                           'dnb_satellite_zenith_angle',
                           'dnb_satellite_azimuth_angle',
                           'dnb_lunar_zenith_angle',
                           'dnb_lunar_azimuth_angle',
                           ])
        self.assertEqual(len(datasets), 6)
        for v in datasets.values():
            self.assertEqual(v.attrs['units'], 'degrees')
            self.assertEqual(v.attrs['rows_per_scan'], 2)
            self.assertEqual(v.attrs['area'].lons.attrs['rows_per_scan'], 2)
            self.assertEqual(v.attrs['area'].lats.attrs['rows_per_scan'], 2)
